from functools import lru_cache

import numpy as np

from centrex_tlf.states import CoupledBasisState, CoupledState

from ..constants import BConstants
from ..wigner import sixj_f, threej_f


@lru_cache(maxsize=int(1e6))
def H_LD(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Calculates the "q-term" that couples states with opposite Omega
    shifting e-parity up and f-parity down in energy
    """
    # All quantum numbers the same, except Omega inverts sign
    J = psi.J
    I1 = psi.I1
    I2 = psi.I2
    F1 = psi.F1
    F = psi.F
    mF = psi.mF
    Omega = -psi.Omega

    amp = constants.q * J * (J + 1) / 2
    ket = CoupledBasisState(
        F,
        mF,
        F1,
        J,
        I1,
        I2,
        Omega=Omega,
        electronic_state=psi.electronic_state,
        P=psi.P,
    )

    return CoupledState([(amp, ket)])


@lru_cache(maxsize=int(1e6))
def H_cp1_Tl(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Calculates the lambda-doubling nuclear spin - rotation term
    """
    # Find the quantum numbers of the input state
    Jp = psi.J
    I1p = psi.I1
    I2p = psi.I2
    F1p = psi.F1
    Fp = psi.F
    mFp = psi.mF
    Omegap = psi.Omega

    # I1, I2, F and mF are the same for both states
    I1 = I1p
    I2 = I2p
    F = Fp
    F1 = F1p
    mF = mFp

    # Omegas are opposite
    Omega = -Omegap

    # Calculate the value of q
    q = Omega

    data = []

    # Loop over possible values of J
    # Need J = |Jp-1| ... |Jp+1|
    for J in np.arange(np.abs(Jp - 1), Jp + 2):
        # Calculate matrix element
        amp = (
            -constants.c1p_Tl
            / 2
            * (-1) ** (J + Jp + F1 + I1 - Omegap)
            * sixj_f(I1, Jp, F1, J, I1, 1)
            * np.sqrt(I1 * (I1 + 1) * (2 * I1 + 1) * (2 * J + 1) * (2 * Jp + 1))
            * (
                (-1) ** (J)
                * np.sqrt(J * (J + 1) * (2 * J + 1))
                * threej_f(J, 1, Jp, 0, q, Omegap)
                * threej_f(J, 1, J, -Omega, q, 0)
                + (-1) ** (Jp)
                * np.sqrt(Jp * (Jp + 1) * (2 * Jp + 1))
                * threej_f(Jp, 1, Jp, 0, q, Omegap)
                * threej_f(J, 1, Jp, -Omega, q, 0)
            )
        )

        basis_state = CoupledBasisState(
            F,
            mF,
            F1,
            J,
            I1,
            I2,
            Omega=Omega,
            electronic_state=psi.electronic_state,
            P=psi.P,
        )

        data.append((amp, basis_state))

    return CoupledState(data)



import numpy as np
import scipy.constants as cst

__all__ = ["doppler_shift", "velocity_to_detuning"]


def doppler_shift(velocity: float, frequency: float = 1.1e15) -> float:
    """
    Doppler a velocity shift

    Args:
        velocity (float): velocity in m/s
        frequency (float, optional): frequency in Hz. Defaults to 1.1e15 Hz, the UV
                                    frequency of the X to B TlF transition.

    Returns:
        float: doppler shifted frequency in Hz
    """
    return frequency * (1 + velocity / cst.c)


def velocity_to_detuning(velocity: float, frequency: float = 1.1e15) -> float:
    """
    Convert a velocity to a detuning based on the doppler shift.

    Args:
        velocity (float): velocity in m/s
        frequency (float, optional): frequency in Hz. Defaults to 1.1e15 Hz, the UV
                                    frequency of the X to B TlF transition.

    Returns:
        float: detuning frequency in radial frequency 2π ⋅ Hz
    """
    return (doppler_shift(velocity, frequency) - frequency) * 2 * np.pi



from typing import overload

import numpy as np
import numpy.typing as npt
import scipy.constants as cst

__all__ = [
    "intensity_to_electric_field",
    "electric_field_to_rabi",
    "power_to_rabi_gaussian_beam",
    "power_to_rabi_gaussian_beam_microwave",
    "rabi_to_electric_field",
    "electric_field_to_intensity",
    "rabi_to_power_gaussian_beam",
    "rabi_to_power_gaussian_beam_microwave",
]


@overload
def intensity_to_electric_field(intensity: float) -> float: ...


@overload
def intensity_to_electric_field(
    intensity: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]: ...


def intensity_to_electric_field(intensity):
    """
    Intensity in W/m^2 to electric field

    Args:
        intensity (float): intensity [W/m^2]

    Returns:
        float: electric field
    """
    return np.sqrt((2 / (cst.c * cst.epsilon_0)) * intensity)


@overload
def electric_field_to_rabi(
    electric_field: float, coupling: float, D: float
) -> float: ...


@overload
def electric_field_to_rabi(
    electric_field: npt.NDArray[np.float_], coupling: float, D: float
) -> npt.NDArray[np.float_]: ...


def electric_field_to_rabi(electric_field, coupling, D):
    """
    Rabi rate from an electric field an coupling strength, with the dipole moment D
    default value set to the X to B transition.

    Args:
        electric_field (float): electric field
        coupling (float): coupling strength
        D (float, optional): Dipole moment.

    Returns:
        float: Rabi rate in rotational frequency [2π ⋅ Hz]
    """
    return electric_field * coupling * D / cst.hbar


@overload
def intensity_to_rabi(intensity: float, coupling: float, D: float) -> float: ...


@overload
def intensity_to_rabi(
    intensity: npt.NDArray[np.float_], coupling: float, D: float
) -> npt.NDArray[np.float_]: ...


def intensity_to_rabi(intensity, coupling, D):
    """
    Rabi rate from an intensity

    Args:
        intensity (float): intensity [W/m^2]
        coupling (float): coupling strength
        D (float): dipole moment

    Returns:
        float: rabi rate
    """
    electric_field = intensity_to_electric_field(intensity)
    rabi = electric_field_to_rabi(electric_field, coupling, D)
    return rabi


def power_to_rabi_rectangular_beam(
    power: float,
    coupling: float,
    wx: float,
    wy: float,
    D: float = 2.6675506e-30,
) -> float:
    """
    Rabi rate from laser power and coupling strength for the X to B TLF transition.

    Args:
        power (float): power [W]
        coupling (float): coupling strength
        wx (float): x width [m]
        wy (float): y width [m]
        D (float, optional): Dipole moment. Defaults to 2.6675506e-30 for the X to B TLF
                            transition.

    Returns:
        float: Rabi rate in rotational frequency [2π ⋅ Hz]
    """
    intensity = power / (wx * wy)

    rabi = intensity_to_rabi(intensity, coupling, D)
    return rabi


def power_to_rabi_gaussian_beam(
    power: float,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = 2.6675506e-30,
) -> float:
    """
    Rabi rate from laser power and coupling strength for the X to B TlF transition.

    Args:
        power (float): power [W]
        coupling (float): coupling strength
        sigma_x (float): x standard deviation [m]
        sigma_y (float): y standard deviation [m]
        D (float, optional): Dipole moment. Defaults to 2.6675506e-30 for the X to B TLF
                            transition.

    Returns:
        float: Rabi rate in rotational frequency [2π ⋅ Hz]
    """
    intensity = power / (2 * np.pi * sigma_x * sigma_y)

    rabi = intensity_to_rabi(intensity, coupling, D)
    return rabi


def power_to_rabi_gaussian_beam_microwave(
    power: float,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = 1.4103753e-29,
) -> float:
    """
    Rabi rate from microwave power and coupling strength for the X to X microwave
    transition

    Args:
        power (float): power [W]
        coupling (float): coupling strength
        sigma_x (float): x standard deviation
        sigma_y (float): y standard deviation
        D (float, optional): Dipole moment. Defaults to 1.4103753e-29 for the X to X
                            TlF transition.

    Returns:
        float: Rabi rate in rotational frequency [2π ⋅ Hz]
    """
    return power_to_rabi_gaussian_beam(
        power=power, coupling=coupling, sigma_x=sigma_x, sigma_y=sigma_y, D=D
    )


def rabi_to_electric_field(rabi: float, coupling: float, D: float) -> float:
    """
    Electric field from a given rabi rate and coupling strength

    Args:
        rabi (float): rabi rate in rotational frequency [2π ⋅ Hz]
        coupling (float): coupling strength
        D (float): Dipole moment

    Returns:
        float: electric field
    """
    return rabi * cst.hbar / (coupling * D)


def electric_field_to_intensity(electric_field: float) -> float:
    """
    Intensity in W/m^2 from a given electric field

    Args:
        electric_field (float): electric field

    Returns:
        float: intensity [W/m^2]
    """
    return 1 / 2 * cst.c * cst.epsilon_0 * electric_field**2


def rabi_to_power_gaussian_beam(
    rabi: float,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = 2.6675506e-30,
) -> float:
    """
    power in W given a rabi rate and couling strength

    Args:
        rabi (float): rabi rate in rotational frequency [2π ⋅ Hz]
        coupling (float): coupling strength
        D (float, optional): dipole moment. Defaults to 2.6675506e-30 for the X to B TlF
                            transition..

    Returns:
        float: power [W]
    """
    electric_field = rabi_to_electric_field(rabi, coupling, D)
    intensity = electric_field_to_intensity(electric_field)
    power = intensity * 2 * np.pi * sigma_x * sigma_y
    return power


def rabi_to_power_gaussian_beam_microwave(
    rabi: float,
    coupling: float,
    sigma_x: float,
    sigma_y: float,
    D: float = 1.4103753e-29,
) -> float:
    """
    power in W given a rabi rate and couling strength

    Args:
        rabi (float): rabi rate in rotational frequency [2π ⋅ Hz]
        coupling (float): coupling strength
        D (float, optional): dipole moment. Defaults to 1.4103753e-29 for the X to X TlF
                            transition.

    Returns:
        float: power [W]
    """
    return rabi_to_power_gaussian_beam(rabi, coupling, sigma_x, sigma_y, D)



from dataclasses import dataclass
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
import sympy as smp

from centrex_tlf import states
from centrex_tlf.transitions import (
    MicrowaveTransition,
    OpticalTransition,
    OpticalTransitionType,
)

from .polarization import Polarization
from .utils import check_transition_coupled_allowed, select_main_states

__all__ = [
    "TransitionSelector",
    "generate_transition_selectors",
    "get_possible_optical_transitions",
]


@dataclass
class TransitionSelector:
    ground: Sequence[states.CoupledState]
    excited: Sequence[states.CoupledState]
    polarizations: Sequence[npt.NDArray[np.complex_]]
    polarization_symbols: List[smp.Symbol]
    Ω: smp.Symbol
    δ: smp.Symbol
    description: Optional[str] = None
    type: Optional[str] = None
    ground_main: Optional[states.CoupledState] = None
    excited_main: Optional[states.CoupledState] = None
    phase_modulation: bool = False

    def __repr__(self) -> str:
        if self.description is None:
            J_g = np.unique([g.largest.J for g in self.ground])[0]
            J_e = np.unique([e.largest.J for e in self.excited])[0]
            return f"TransitionSelector(J={J_g} -> J={J_e})"
        else:
            return f"TransitionSelector({self.description})"


def generate_transition_selectors(
    transitions: Sequence[Union[OpticalTransition, MicrowaveTransition]],
    polarizations: Sequence[Sequence[Polarization]],
    ground_mains: Optional[Sequence[states.CoupledState]] = None,
    excited_mains: Optional[Sequence[states.CoupledState]] = None,
    phase_modulations: Optional[Sequence[bool]] = None,
) -> List[TransitionSelector]:
    """
    Generate a list of TransitionSelectors from Transition(s) and Polarization(s).

    Args:
        transitions (Sequence[Union[OpticalTransition, MicrowaveTransition]]):
                                                    transitions to include in the system
        polarizations (Sequence[Sequence[Polarization]]): polarization, list of
                                                        polarizations per transition.
        ground_mains (Optional[Sequence[States]]): Sequence of a main ground state to
                                                    use per transition
        excited_mains (Optional[Sequence[States]]): Sequence of a a main excited state
                                                    to use per transition
        excited_mains:

    Returns:
        List[TransitionSelector]: List of TransitionSelectors
    """
    transition_selectors = []

    for idt, (transition, polarization) in enumerate(zip(transitions, polarizations)):
        ground_states_approx_qn_select = states.QuantumSelector(
            J=transition.J_ground,
            electronic=transition.electronic_ground,
            P=transition.P_ground,
            Ω=transition.Ω_ground,
        )
        ground_states_approx = list(
            [
                1 * s
                for s in states.generate_coupled_states_X(
                    ground_states_approx_qn_select
                )
            ]
        )

        if isinstance(transition, OpticalTransition):
            excited_states_approx = [
                1 * s
                for s in states.generate_coupled_states_B(transition.qn_select_excited)
            ]
        elif isinstance(transition, MicrowaveTransition):
            excited_states_approx = [
                1 * s
                for s in states.generate_coupled_states_X(transition.qn_select_excited)
            ]

        if phase_modulations is None:
            phase_modulation = False
        else:
            phase_modulation = phase_modulations[idt]
        if ground_mains is None:
            ground_main, excited_main = select_main_states(
                ground_states_approx, excited_states_approx, polarization[0].vector
            )

        transition_selectors.append(
            TransitionSelector(
                ground=ground_states_approx,
                excited=excited_states_approx,
                polarizations=[p.vector for p in polarization],
                polarization_symbols=[
                    smp.Symbol(f"P{p.name}{idt}") for p in polarization
                ],
                Ω=smp.Symbol(f"Ω{idt}", complex=True),
                δ=smp.Symbol(f"δ{idt}"),
                description=transition.name,
                ground_main=ground_main if ground_mains is None else ground_mains[idt],
                excited_main=excited_main
                if excited_mains is None
                else excited_mains[idt],
                phase_modulation=phase_modulation,
            )
        )
    return transition_selectors


def get_possible_optical_transitions(
    ground_state: states.CoupledBasisState,
    transition_types: Optional[Sequence[OpticalTransitionType]] = None,
):
    J = ground_state.J
    # F1 = ground_state.F1
    # F = ground_state.F
    I1 = ground_state.I1
    I2 = ground_state.I2

    if transition_types is None:
        transition_types = [t for t in OpticalTransitionType]

    transitions = []
    for transition_type in transition_types:
        ΔJ = transition_type.value
        J_excited = J + ΔJ
        _transitions = [
            OpticalTransition(transition_type, J, F1, F)
            for F1 in np.arange(np.abs(J_excited - I1), J_excited + I1 + 1)
            for F in np.arange(np.abs(F1 - I2), F1 + I2 + 1, dtype=int)
        ]
        _transitions = [
            t
            for t in _transitions
            if check_transition_coupled_allowed(ground_state, t.excited_states[0])
        ]
        transitions.append(_transitions)
    return transitions



from typing import Any, Sequence, Union

import numpy as np
import numpy.typing as npt

from centrex_tlf.states import BasisState

__all__ = ["generate_transform_matrix"]


def generate_transform_matrix(
    basis1: Union[Sequence[BasisState], npt.NDArray[Any]],
    basis2: Union[Sequence[BasisState], npt.NDArray[Any]],
) -> npt.NDArray[np.complex128]:
    """
    Function that generates a transform matrix that takes Hamiltonian expressed
    in basis1 to basis2: H_2 = S.conj().T @ H_1 @ S

    inputs:
    basis1 = list of basis states that defines basis1
    basis2 = list of basis states that defines basis2
    progress = boolean to display tqdm progress bar

    returns:
    S = transformation matrix that takes Hamiltonian (or any operator) from
    basis1 to basis2
    """

    # Check that the two bases have the same dimension
    assert len(basis1) == len(basis2), "Bases don't have the same dimension"

    # Initialize S
    S = np.zeros((len(basis1), len(basis1)), dtype=complex)

    # Loop over the bases and calculate inner products
    for i, state1 in enumerate(basis1):
        for j, state2 in enumerate(basis2):
            S[i, j] = state1 @ state2

    # Return the transform matrix
    return S



import copy
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt
from scipy import constants as cst

from .find_states import QuantumSelector
from .states import CoupledState
from .utils_compact import compact_QN_coupled_indices

__all__ = [
    "thermal_population",
    "J_levels",
    "generate_thermal_population_states",
    "generate_population_states",
]


def thermal_population(J: int, T: float, B: float = 6.66733e9, n: int = 100) -> float:
    """calculate the thermal population of a given J sublevel

    Args:
        J (int): rotational level
        T (float): temperature [Kelvin]
        B (float, optional): rotational constant. Defaults to 6.66733e9.
        n (int, optional): number of rotational levels to normalize with.
                            Defaults to 100.

    Returns:
        float: relative population in a rotational sublevel
    """
    c = 2 * np.pi * cst.hbar * B / (cst.k * T)

    def a(J):
        return -c * J * (J + 1)

    Z = np.sum([J_levels(i) * np.exp(a(i)) for i in range(n)])
    return J_levels(J) * np.exp(a(J)) / Z


def J_levels(J: int) -> int:
    """calculate the number of hyperfine sublevels per J rotational level

    Args:
        J (int): rotational level

    Returns:
        int: number of levels
    """
    return 4 * (2 * J + 1)


def generate_thermal_population_states(
    states_to_fill: Union[Sequence[QuantumSelector], QuantumSelector],
    states: Sequence[CoupledState],
    T: float,
    qn_compact: Optional[Union[Sequence[QuantumSelector], QuantumSelector]] = None,
) -> npt.NDArray[np.complex_]:
    """Generate a thermal distrubtion over the states specified in
    states_to_fill, a QuantumSelector or list of Quantumselectors

    Args:
        states_to_fill (QuantumSelector): Quantumselector specifying states to
        fill
        states (list, np.ndarray): all states used in simulation
        T (float): temperature in Kelvin
        qn_compact (Optional[Union[Sequence[QuantumSelector], QuantumSelector]]):
            defaults to None. Quantum selector to specify which states are compacted
            in the OBE system.

    Returns:
        np.ndarray: density matrix with trace normalized to 1
    """
    # branch for single QuantumSelector use
    if isinstance(states_to_fill, QuantumSelector):
        # get all involved Js
        Js = states_to_fill.J
        assert Js is not None, (
            "states_to_fill needs rotational levels defined to generate the thermal "
            "population density"
        )
        # check if J was a list
        _Js = (
            np.array([Js])
            if not isinstance(Js, (np.ndarray, list, tuple, Sequence))
            else np.array(Js)
        )
        # get indices of states to fill
        indices_to_fill = states_to_fill.get_indices(states)
    # branch for multiple QuantumSelectors use
    elif isinstance(states_to_fill, (list, np.ndarray, tuple)):
        assert isinstance(states_to_fill[0], QuantumSelector), (
            "need to supply a sequence of QuantumSelectors, not "
            f"{type(states_to_fill[0])}"
        )
        # get all involved Js
        _Js = np.array([], dtype=np.int_)
        for stf in states_to_fill:
            J = stf.J
            # check if J was a list
            if J is not None:
                _J = (
                    [J] if not isinstance(J, (np.ndarray, list, tuple, Sequence)) else J
                )
                _Js = np.append(_Js, _J)
        assert (
            len(_Js) != 0
        ), "requires states_to_fill with QuantumSelectors with rotation levels defined"
        # get indices of states to fill
        indices_to_fill = np.array([], dtype=np.int_)
        for stf in states_to_fill:
            indices_to_fill = np.append(
                indices_to_fill, stf.get_indices(states)
            ).astype(int)

    # remove duplicates from Js and indices_to_fill
    _Js = np.unique(_Js)
    indices_to_fill = np.unique(indices_to_fill)

    # thermal population per hyperfine level for each involved J
    thermal_populations = dict(
        [(Ji, thermal_population(Ji, T) / J_levels(Ji)) for Ji in _Js]
    )
    # generate an empty density matrix
    density = np.zeros(len(states), dtype=complex)
    # fill the density matrix
    for idd in indices_to_fill:
        state = states[idd].largest
        thermal_pop = thermal_populations[state.J]
        density[idd] = thermal_pop

    if qn_compact is not None:
        states_compact = copy.copy(states)
        if isinstance(qn_compact, (list, tuple, np.ndarray, Sequence)):
            for qnc in qn_compact:
                indices_compact = qnc.get_indices(states_compact)
                pop_compact = density[indices_compact].sum()
                density = np.array(
                    [di for idd, di in enumerate(density) if idd not in indices_compact]
                )
                density[indices_compact[0]] = pop_compact
                states_compact = compact_QN_coupled_indices(
                    states_compact, indices_compact
                )
        elif isinstance(qn_compact, QuantumSelector):
            indices_compact = qn_compact.get_indices(states_compact)
            pop_compact = density[indices_compact].sum()
            density = np.array(
                [di for idd, di in enumerate(density) if idd not in indices_compact]
            )
            density[indices_compact[0]] = pop_compact
            states_compact = compact_QN_coupled_indices(states_compact, indices_compact)
        else:
            raise AssertionError(
                "qn_compact required to be a QuantumsSelector or a Sequence of "
                f"QuantumSelectors, not {type(qn_compact)}"
            )

    density = np.eye(len(density), dtype=np.complex_) * density

    # normalize the trace to 1 and return the density matrix
    return density / np.trace(density)


def generate_population_states(
    states: Union[List[int], npt.NDArray[np.int_]], levels: int
):
    """generate a uniform population distribution with population in the
    specified states

    Args:
        states (list, np.ndarray): indices to put population into
        levels (int): total number of levels

    Returns:
        np.ndarray: density matrix
    """
    density = np.zeros([levels, levels], dtype=complex)
    for state in states:
        density[state, state] = 1
    return density / np.trace(density)



from typing import Literal, Tuple, Union, overload

import numpy as np
import numpy.typing as npt
import sympy as smp

from .utils import generate_density_matrix_symbolic

__all__ = ["generate_system_of_equations_symbolic"]


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    C_array: npt.NDArray[np.float_],
    fast: bool,
    split_output: Literal[False],
) -> smp.matrices.dense.MutableDenseMatrix:
    ...


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    C_array: npt.NDArray[np.float_],
    fast: bool,
) -> smp.matrices.dense.MutableDenseMatrix:
    ...


@overload
def generate_system_of_equations_symbolic(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    C_array: npt.NDArray[np.float_],
    fast: bool,
    split_output: Literal[True],
) -> Tuple[
    smp.matrices.dense.MutableDenseMatrix, smp.matrices.dense.MutableDenseMatrix
]:
    ...


def generate_system_of_equations_symbolic(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    C_array: npt.NDArray[np.float_],
    fast: bool = False,
    split_output: bool = False,
) -> Union[
    smp.matrices.dense.MutableDenseMatrix,
    Tuple[smp.matrices.dense.MutableDenseMatrix, smp.matrices.dense.MutableDenseMatrix],
]:
    n_states = hamiltonian.shape[0]
    ρ = generate_density_matrix_symbolic(n_states)
    C_conj_array = np.einsum("ijk->ikj", C_array.conj())

    matrix_mult_sum = smp.zeros(n_states, n_states)
    if fast:
        # C_array is an array of 2D arrays, where each 2D array only has one
        # entry, i.e. don't have to do the full matrix multiplication each
        # time for C@ρ@Cᶜ, i.e. using manual spare matrix multiplication
        for C, Cᶜ in zip(C_array, C_conj_array):
            idC = np.nonzero(C)
            idCᶜ = np.nonzero(Cᶜ)
            val = C[idC][0] * Cᶜ[idCᶜ][0] * ρ[idC[-1], idCᶜ[0]][0]
            matrix_mult_sum[idC[0][0], idCᶜ[-1][0]] += val

    else:
        for idx in range(C_array.shape[0]):
            matrix_mult_sum[:, :] += C_array[idx] @ ρ @ C_conj_array[idx]

    Cprecalc = np.einsum("ijk,ikl", C_conj_array, C_array)

    a = -0.5 * (Cprecalc @ ρ + ρ @ Cprecalc)
    b = -1j * (hamiltonian @ ρ - ρ @ hamiltonian)

    if split_output:
        return b, matrix_mult_sum + a
    else:
        system = smp.zeros(n_states, n_states)
        system += matrix_mult_sum + a + b
        return system



from typing import List, Optional, Sequence

import numpy as np
import numpy.typing as npt
import pandas as pd

from centrex_tlf import states

from .matrix_elements import calculate_ED_ME_mixed_state

__all__ = ["calculate_br", "generate_br_dataframe"]


def calculate_br(
    excited_state: states.CoupledState,
    ground_states: Sequence[states.CoupledState],
    tol: float = 1e-3,
) -> npt.NDArray[np.float_]:
    # matrix elements between the excited state and the ground states
    MEs = np.zeros((len(ground_states)), dtype=np.complex_)

    for idg, ground_state in enumerate(ground_states):
        MEs[idg] = calculate_ED_ME_mixed_state(
            ground_state.remove_small_components(tol=tol),
            excited_state.remove_small_components(tol=tol),
        )

    # Calculate branching ratios
    BRs = np.abs(MEs) ** 2 / (np.sum(np.abs(MEs) ** 2)).astype(np.float_)
    return BRs


def generate_br_dataframe(
    ground_states: Sequence[states.CoupledState],
    excited_states: Sequence[states.CoupledState],
    group_ground: Optional[str] = None,
    group_excited: bool = True,
    remove_zeros: bool = True,
    tolerance: float = 1e-3,
) -> pd.DataFrame:
    """
    Generate a pandas DataFrame of branching ratios given a set of ground states and
    excited states.

    Args:
        ground_states (Sequence[states.State]): superpositions of ground state
                                                CoupledBasisStates
        excited_states (Sequence[states.State]): superpositions of excited state
                                                CoupledBasisStates
        group_ground (Optional[str], optional): Group ground states with "J" or "mF".
                                                Defaults to None.
        group_excited (bool, optional): Group all excited states in a single column.
                                        Defaults to True.
        remove_zeros (bool, optional): Remove states with zero branching. Defaults to
                                        True.

    Raises:
        AssertionError: _description_

    Returns:
        _type_: _description_
    """
    br: List[npt.NDArray[np.float_]] = []
    for es in excited_states:
        br.append(calculate_br(es, ground_states, tolerance))

    brs = np.sum(br, axis=0)

    if group_ground is not None:
        if group_ground == "J":
            J_unique = np.unique([s.largest.J for s in ground_states])
            indices_group = [
                states.QuantumSelector(
                    J=Ji, electronic=states.ElectronicState.X
                ).get_indices(ground_states)
                for Ji in J_unique
            ]
            data = {"states": [f"|X, J = {Ji}>" for Ji in J_unique]}
        elif group_ground == "mF":
            mF_selectors = np.unique(
                [(s.largest.J, s.largest.F1, s.largest.F) for s in ground_states],  # type: ignore # noqa: 203
                axis=0,
            )
            indices_group = [
                states.QuantumSelector(
                    J=J, F1=F1, F=F, electronic=states.ElectronicState.X
                ).get_indices(ground_states)
                for J, F1, F in mF_selectors
            ]
            data = {
                "states": [
                    [ground_states[idx] for idx in ind][0].largest.state_string_custom(
                        ["electronic", "J", "F1", "F"]
                    )
                    for ind in indices_group
                ]
            }
        else:
            raise AssertionError("group_ground not equal to either J or mF")
        br = [np.array([bri[ind].sum() for ind in indices_group]) for bri in br]
        brs = np.sum(br, axis=0)
        if remove_zeros:
            m = brs != 0
        else:
            m = np.ones(len(brs), dtype=bool)
        data["states"] = np.asarray(data["states"])[m]
    else:
        if remove_zeros:
            m = brs != 0
        else:
            m = np.ones(len(brs), dtype=bool)
        data = {
            "states": [
                qn.largest.state_string_custom(  # type: ignore
                    ["electronic", "J", "F1", "F", "mF"]
                )
                for qn in [s for ids, s in enumerate(ground_states) if m[ids]]
            ]
        }

    br_dataframe = pd.DataFrame(data=data)
    if group_excited:
        J_unique = np.unique([s.largest.J for s in excited_states])
        F1_unique: npt.NDArray[np.float_] = np.unique(
            [s.largest.F1 for s in excited_states]  # type: ignore
        )
        F_unique: npt.NDArray[np.int_] = np.unique(
            [s.largest.F for s in excited_states]  # type: ignore
        )
        quantum_selectors = [
            states.QuantumSelector(
                J=Ji, F1=F1i, F=Fi, electronic=states.ElectronicState.B
            )
            for Ji in J_unique
            for F1i in F1_unique
            for Fi in F_unique
        ]
        indices_group = [qs.get_indices(excited_states) for qs in quantum_selectors]
        indices_group = [ind for ind in indices_group if len(ind) > 0]
        for ind in indices_group:
            s = excited_states[ind[0]].largest
            bri = np.sum([br[i] for i in ind], axis=0)
            bri /= np.sum(bri)
            br_dataframe[
                s.state_string_custom(["electronic", "J", "F1", "F"])  # type: ignore
            ] = bri[m]
    else:
        for idb, brv in enumerate(br):
            br_dataframe[
                excited_states[idb].largest.state_string_custom(  # type: ignore
                    ["electronic", "J", "F1", "F", "mF"]
                )
            ] = np.asarray(brv)[m] / np.sum(brs)
    return br_dataframe.set_index("states")



import copy
from typing import List, Optional, Sequence, Union

import numpy as np
import numpy.typing as npt

from centrex_tlf import states

from .branching import calculate_br
from .utils_compact import compact_C_array_indices

__all__ = ["collapse_matrices"]


def collapse_matrices(
    QN: Sequence[states.CoupledState],
    ground_states: Sequence[states.CoupledState],
    excited_states: Sequence[states.CoupledState],
    gamma: float = 1,
    tol: float = 1e-4,
    qn_compact: Optional[
        Union[states.QuantumSelector, Sequence[states.QuantumSelector]]
    ] = None,
) -> npt.NDArray[np.float_]:
    """
    Function that generates the collapse matrix for given ground and excited states

    inputs:
    QN = list of states that defines the basis for the calculation
    ground_states = list of ground states that are coupled to the excited states
    excited_states = list of excited states that are coupled to the ground states
    gamma = decay rate of excited states
    tol = couplings smaller than tol/sqrt(gamma) are set to zero to speed up computation
    qn_compact = list of QuantumSelectors or lists of QuantumSelectors with each
                QuantumSelector containing the quantum numbers to compact into a
                single state. Defaults to None.

    outputs:
    C_list = array of collapse matrices
    """
    # Initialize list of collapse matrices
    C_list: List[npt.NDArray[np.float_]] = []

    # Start looping over ground and excited states
    for excited_state in excited_states:
        j = QN.index(excited_state)
        BRs = calculate_br(excited_state, ground_states)
        if np.sum(BRs) > 1:
            print(f"Warning: Branching ratio sum > 1, difference = {np.sum(BRs)-1:.2e}")
        for ground_state, BR in zip(ground_states, BRs):
            i = QN.index(ground_state)

            if np.sqrt(BR) > tol:
                # Initialize the coupling matrix
                H = np.zeros((len(QN), len(QN)), dtype=np.float64)
                H[i, j] = np.sqrt(BR * gamma)

                C_list.append(H)

    C_array = np.array(C_list)

    if qn_compact:
        if isinstance(qn_compact, states.QuantumSelector):
            qn_compact = [qn_compact]
        QN_compact = copy.deepcopy(QN)
        for qnc in qn_compact:
            indices_compact = states.get_indices_quantumnumbers(qnc, QN_compact)
            QN_compact = states.compact_QN_coupled_indices(QN_compact, indices_compact)
            C_array = compact_C_array_indices(C_array, gamma, indices_compact)
    return C_array



from dataclasses import dataclass
from typing import Sequence, Union, cast

import numpy as np
import numpy.typing as npt
import pandas as pd

from centrex_tlf import states
from centrex_tlf.states.states import CoupledBasisState

from .matrix_elements import calculate_ED_ME_mixed_state
from .utils import ΔmF_allowed, assert_transition_coupled_allowed, select_main_states

__all__ = [
    "generate_coupling_matrix",
    "generate_coupling_field",
    "generate_coupling_field_automatic",
    "CouplingFields",
    "CouplingField",
    "generate_coupling_dataframe",
]


def generate_coupling_matrix(
    QN: Sequence[states.CoupledState],
    ground_states: Sequence[states.CoupledState],
    excited_states: Sequence[states.CoupledState],
    pol_vec: npt.NDArray[np.complex_] = np.array([0.0, 0.0, 1.0], dtype=np.complex_),
    reduced: bool = False,
    normalize_pol: bool = True,
) -> npt.NDArray[np.complex_]:
    """generate optical coupling matrix for given ground and excited states
    Checks if couplings are already pre-cached, otherwise falls back to
    calculate_coupling_matrix.

    Args:
        QN (list): list of basis states
        ground_states (list): list of ground states coupling to excited states
        excited_states (list): list of excited states
        pol_vec (np.ndarray): polarization vector. Defaults to
                                        np.array([0,0,1]).
        reduced (bool, optional): [description]. Defaults to False.
        normalize_pol (bool, optional): Normalize the polarization vector.
                                        Defaults to True.

    Returns:
        np.ndarray: optical coupling matrix
    """
    assert isinstance(QN, list), "QN required to be of type list"

    H = np.zeros((len(QN), len(QN)), dtype=complex)

    # start looping over ground and excited states
    for ground_state in ground_states:
        i = QN.index(ground_state)
        for excited_state in excited_states:
            j = QN.index(excited_state)

            # calculate matrix element and add it to the Hamiltonian
            H[i, j] = calculate_ED_ME_mixed_state(
                excited_state,
                ground_state,
                pol_vec=pol_vec,
                reduced=reduced,
                normalize_pol=normalize_pol,
            )

            # make H hermitian
    H = H + H.conj().T

    return H


@dataclass
class CouplingField:
    polarization: npt.NDArray[np.complex_]
    field: npt.NDArray[np.complex_]


@dataclass
class CouplingFields:
    ground_main: states.CoupledState
    excited_main: states.CoupledState
    main_coupling: complex
    ground_states: Sequence[states.CoupledState]
    excited_states: Sequence[states.CoupledState]
    fields: Sequence[CouplingField]

    def __repr__(self):
        gs = self.ground_main.largest
        es = self.excited_main.largest
        gs_str = gs.state_string_custom(["electronic", "J", "F1", "F", "mF", "P", "Ω"])
        es_str = es.state_string_custom(["electronic", "J", "F1", "F", "mF", "P", "Ω"])
        return (
            f"CouplingFields(ground_main={gs_str},"
            f" excited_main={es_str},"
            f" main_coupling={self.main_coupling:.2e}"
        )


def _generate_coupling_dataframe(
    field: CouplingField, states_list: Sequence[states.CoupledState]
) -> pd.DataFrame:
    indices = np.nonzero(np.triu(field.field))
    ground_states = []
    excited_states = []
    couplings = []
    for idx, idy in zip(*indices):
        gs = states_list[idx].largest.state_string_custom(
            ["electronic", "J", "F1", "F", "mF"]
        )
        es = states_list[idy].largest.state_string_custom(
            ["electronic", "J", "F1", "F", "mF"]
        )
        ground_states.append(gs)
        excited_states.append(es)
        couplings.append(field.field[idx, idy])

    data = {"ground": ground_states, "excited": excited_states, "couplings": couplings}
    return pd.DataFrame(data)


def generate_coupling_dataframe(
    fields: CouplingFields, states_list: Sequence[states.CoupledState]
) -> Sequence[pd.DataFrame]:
    """
    Generate a list of pandas DataFrames with the non-zero couplings between states
    listed for each separate CouplingField input

    Args:
        fields (CouplingFields): coupling fields for a given transitions, with one for
        each polarization
        states_list (Sequence[states.State]): states involved in the system

    Returns:
        Sequence[pd.DataFrame]: list of DataFrames with non-zero couplings
    """
    dfs = []
    for field in fields.fields:
        dfs.append(_generate_coupling_dataframe(field, states_list))
    return dfs


def generate_coupling_field(
    ground_main_approx: states.CoupledState,
    excited_main_approx: states.CoupledState,
    ground_states_approx: Union[
        Sequence[states.CoupledState], Sequence[states.CoupledBasisState]
    ],
    excited_states_approx: Union[
        Sequence[states.CoupledState], Sequence[states.CoupledBasisState]
    ],
    QN_basis: Union[Sequence[states.CoupledState], Sequence[states.CoupledBasisState]],
    H_rot: npt.NDArray[np.complex_],
    QN: Sequence[states.CoupledState],
    V_ref: npt.NDArray[np.complex_],
    pol_main: npt.NDArray[np.complex_] = np.array([0, 0, 1], dtype=np.complex_),
    pol_vecs: Sequence[npt.NDArray[np.complex_]] = [],
    relative_coupling: float = 1e-3,
    absolute_coupling: float = 1e-6,
    normalize_pol: bool = True,
) -> CouplingFields:
    assert isinstance(
        pol_main, np.ndarray
    ), "supply a Sequence of np.ndarrays with dtype np.complex_ for pol_vecs"
    assert isinstance(
        pol_vecs[0], np.ndarray
    ), "supply a Sequence of np.ndarrays with dtype np.complex_ for pol_vecs"
    if not np.issubdtype(pol_main.dtype, np.complex_):
        pol_main.astype(np.complex_)
    if not np.issubdtype(pol_vecs[0].dtype, np.complex_):
        pol_vecs = [pol.astype(np.complex_) for pol in pol_vecs]

    _ground_states_approx: Sequence[states.CoupledState]
    _excited_states_approx: Sequence[states.CoupledState]
    _QN_basis: Sequence[states.CoupledState]

    if isinstance(ground_states_approx[0], CoupledBasisState):
        ground_states_approx = cast(Sequence[CoupledBasisState], ground_states_approx)
        _ground_states_approx = states.states.basisstate_to_state_list(
            ground_states_approx
        )
    else:
        _ground_states_approx = cast(
            Sequence[states.CoupledState], ground_states_approx
        )

    if isinstance(excited_states_approx[0], CoupledBasisState):
        excited_states_approx = cast(Sequence[CoupledBasisState], excited_states_approx)
        _excited_states_approx = states.states.basisstate_to_state_list(
            excited_states_approx
        )
    else:
        _excited_states_approx = cast(
            Sequence[states.CoupledState], excited_states_approx
        )

    if isinstance(QN_basis[0], CoupledBasisState):
        QN_basis = cast(Sequence[CoupledBasisState], QN_basis)
        _QN_basis = states.states.basisstate_to_state_list(QN_basis)
    else:
        _QN_basis = cast(Sequence[states.CoupledState], QN_basis)

    ground_states = states.find_exact_states(
        _ground_states_approx, _QN_basis, QN, H_rot, V_ref=V_ref
    )
    excited_states = states.find_exact_states(
        _excited_states_approx, _QN_basis, QN, H_rot, V_ref=V_ref
    )
    ground_main = states.find_exact_states(
        [ground_main_approx], _QN_basis, QN, H_rot, V_ref=V_ref
    )[0]
    excited_main = states.find_exact_states(
        [excited_main_approx], _QN_basis, QN, H_rot, V_ref=V_ref
    )[0]

    states.check_approx_state_exact_state(ground_main_approx, ground_main)
    states.check_approx_state_exact_state(excited_main_approx, excited_main)
    ME_main = calculate_ED_ME_mixed_state(
        excited_main,
        ground_main,
        pol_vec=np.asarray(pol_main, dtype=np.complex_),
        normalize_pol=normalize_pol,
    )

    assert (
        ME_main != 0
    ), f"main coupling element for {ground_main_approx} -> {excited_main_approx} is zero"

    _ground_main = cast(CoupledBasisState, ground_main.largest)
    _excited_main = cast(CoupledBasisState, excited_main.largest)

    assert_transition_coupled_allowed(
        _ground_main, _excited_main, ΔmF_allowed(pol_main)
    )

    couplings = []
    for pol in pol_vecs:
        coupling = generate_coupling_matrix(
            QN,
            ground_states,
            excited_states,
            pol_vec=pol,
            reduced=False,
            normalize_pol=normalize_pol,
        )
        if normalize_pol:
            pol = pol.copy() / np.linalg.norm(pol)

        coupling[np.abs(coupling) < relative_coupling * np.max(np.abs(coupling))] = 0
        coupling[np.abs(coupling) < absolute_coupling] = 0
        couplings.append(CouplingField(polarization=pol, field=coupling))
    return CouplingFields(
        ground_main, excited_main, ME_main, ground_states, excited_states, couplings
    )


def generate_coupling_field_automatic(
    ground_states_approx: Union[
        Sequence[states.CoupledState],
        Sequence[states.CoupledBasisState],
        Sequence[states.UncoupledBasisState],
    ],
    excited_states_approx: Union[
        Sequence[states.CoupledState],
        Sequence[states.CoupledBasisState],
        Sequence[states.UncoupledBasisState],
    ],
    QN_basis: Union[
        Sequence[states.CoupledState],
        Sequence[states.CoupledBasisState],
        Sequence[states.UncoupledBasisState],
    ],
    H_rot: npt.NDArray[np.complex_],
    QN: Sequence[states.CoupledState],
    V_ref: npt.NDArray[np.complex_],
    pol_vecs: Sequence[npt.NDArray[np.complex_]],
    relative_coupling: float = 1e-3,
    absolute_coupling: float = 1e-6,
    normalize_pol: bool = True,
) -> CouplingFields:
    """Calculate the coupling fields for a transition for one or multiple
    polarizations.

    Args:
        ground_states_approx (list): list of approximate ground states
        excited_states_approx (list): list of approximate excited states
        QN_basis (Sequence[states.State]): Sequence of States the H_rot was constructed
                                            from
        H_rot (np.ndarray): System hamiltonian in the rotational frame
        QN (list): list of states in the system
        V_ref ([type]): [description]
        pol_vec (list): list of polarizations.
        relative_coupling (float): minimum relative coupling, set
                                            smaller coupling to zero.
                                            Defaults to 1e-3.
        absolute_coupling (float): minimum absolute coupling, set
                                            smaller couplings to zero.
                                            Defaults to 1e-6.

    Returns:
        dictionary: CouplingFields dataclass with the coupling information.
                    Attributes:
                        ground_main: main ground state
                        excited_main: main excited state
                        main_coupling: coupling strenght between main_ground
                                        and main_excited
                        ground_states: ground states in coupling
                        excited_states: excited_states in coupling
                        fields: list of CouplingField dataclasses, one for each
                                polarization, containing the polarization and coupling
                                field
    """
    assert isinstance(
        pol_vecs[0], np.ndarray
    ), "supply a Sequence of np.ndarrays with dtype np.float_ for pol_vecs"

    _ground_states_approx: Sequence[states.CoupledState]
    _excited_states_approx: Sequence[states.CoupledState]
    _QN_basis: Sequence[states.CoupledState]

    if isinstance(ground_states_approx[0], CoupledBasisState):
        ground_states_approx = cast(Sequence[CoupledBasisState], ground_states_approx)
        _ground_states_approx = states.states.basisstate_to_state_list(
            ground_states_approx
        )
    else:
        _ground_states_approx = cast(
            Sequence[states.CoupledState], ground_states_approx
        )

    if isinstance(excited_states_approx[0], CoupledBasisState):
        excited_states_approx = cast(Sequence[CoupledBasisState], excited_states_approx)
        _excited_states_approx = states.states.basisstate_to_state_list(
            excited_states_approx
        )
    else:
        _excited_states_approx = cast(
            Sequence[states.CoupledState], excited_states_approx
        )

    if isinstance(QN_basis[0], CoupledBasisState):
        QN_basis = cast(Sequence[CoupledBasisState], QN_basis)
        _QN_basis = states.states.basisstate_to_state_list(QN_basis)
    else:
        _QN_basis = cast(Sequence[states.CoupledState], QN_basis)

    pol_main = pol_vecs[0]
    ground_main_approx, excited_main_approx = select_main_states(
        _ground_states_approx, _excited_states_approx, pol_main
    )
    return generate_coupling_field(
        ground_main_approx=ground_main_approx,
        excited_main_approx=excited_main_approx,
        ground_states_approx=_ground_states_approx,
        excited_states_approx=_excited_states_approx,
        QN_basis=_QN_basis,
        H_rot=H_rot,
        QN=QN,
        V_ref=V_ref,
        pol_main=pol_main,
        pol_vecs=pol_vecs,
        relative_coupling=relative_coupling,
        absolute_coupling=absolute_coupling,
        normalize_pol=normalize_pol,
    )



from dataclasses import dataclass
from enum import IntEnum
from typing import Sequence

import sympy

import centrex_tlf.states as states

__all__: list[str] = [
    "OpticalTransitionType",
    "MicrowaveTransition",
    "OpticalTransition",
]


class OpticalTransitionType(IntEnum):
    O = -2  # noqa: E741
    P = -1
    Q = 0
    R = +1
    S = +2


@dataclass
class MicrowaveTransition:
    J_ground: int
    J_excited: int
    electronic_ground: states.ElectronicState = states.ElectronicState.X
    electronic_excited: states.ElectronicState = states.ElectronicState.X

    def __repr__(self) -> str:
        return f"MicrowaveTransition({self.name})"

    @property
    def name(self) -> str:
        return f"J={self.J_ground} -> J={self.J_excited}"

    @property
    def Ω_ground(self) -> int:
        return 0

    @property
    def Ω_excited(self) -> int:
        return 0

    @property
    def P_ground(self) -> int:
        return (-1) ** self.J_ground

    @property
    def P_excited(self) -> int:
        return (-1) ** self.J_excited

    @property
    def qn_select_ground(self) -> states.QuantumSelector:
        return states.QuantumSelector(
            J=self.J_ground, electronic=self.electronic_ground, Ω=self.Ω_ground
        )

    @property
    def qn_select_excited(self) -> states.QuantumSelector:
        return states.QuantumSelector(
            J=self.J_excited, electronic=self.electronic_excited, Ω=self.Ω_excited
        )


@dataclass
class OpticalTransition:
    t: OpticalTransitionType
    J_ground: int
    F1: float
    F: int
    electronic_ground: states.ElectronicState = states.ElectronicState.X
    electronic_excited: states.ElectronicState = states.ElectronicState.B

    def __repr__(self) -> str:
        return f"OpticalTransition({self.name})"

    @property
    def name(self) -> str:
        F1 = sympy.S(str(self.F1), rational=True)
        return f"{self.t.name}({self.J_ground}) F1'={F1} F'={int(self.F)}"

    @property
    def J_excited(self) -> int:
        return self.J_ground + self.t.value

    @property
    def P_excited(self) -> int:
        return self.P_ground * -1

    @property
    def P_ground(self) -> int:
        return (-1) ** self.J_ground

    @property
    def Ω_excited(self) -> int:
        return 1

    @property
    def Ω_ground(self) -> int:
        return 0

    @property
    def qn_select_ground(self) -> states.QuantumSelector:
        return states.QuantumSelector(
            J=self.J_ground, electronic=self.electronic_ground, Ω=self.Ω_ground
        )

    @property
    def qn_select_excited(self) -> states.QuantumSelector:
        return states.QuantumSelector(
            J=self.J_excited,
            F1=self.F1,
            F=self.F,
            electronic=self.electronic_excited,
            P=self.P_excited,
            Ω=self.Ω_excited,
        )

    @property
    def ground_states(self) -> Sequence[states.CoupledBasisState]:
        return states.generate_coupled_states_X(self.qn_select_ground)

    @property
    def excited_states(self) -> Sequence[states.CoupledBasisState]:
        return states.generate_coupled_states_B(self.qn_select_excited)



from typing import List, Sequence, Tuple, Union, cast

import numpy as np
import numpy.typing as npt

from centrex_tlf import states
from centrex_tlf.states.states import CoupledBasisState

from .polarization import (
    polarization_X,
    polarization_Y,
    polarization_Z,
    polarization_σm,
    polarization_σp,
)

__all__: List[str] = []


def check_transition_coupled_allowed(
    ground: states.CoupledBasisState, excited: states.CoupledBasisState
):
    """
    Check whether the transition between the two states is allowed based on the quantum
    numbers

    Args:
        ground (states.CoupledBasisState): ground CoupledBasisState
        excited (states.CoupledBasisState): excited CoupledBasisState
    """
    assert ground.P is not None, "parity is required to be set for ground"
    assert excited.P is not None, "parity is required to be set for excited"

    ΔF = int(excited.F - ground.F)
    ΔP = int(excited.P - ground.P)

    flag_ΔP = abs(ΔP) != 2
    flag_ΔF = abs(ΔF) > 1

    return not (flag_ΔP | flag_ΔF)


def check_transition_coupled_allowed_polarization(
    ground_state: states.CoupledBasisState,
    excited_state: states.CoupledBasisState,
    ΔmF_allowed: int,
    return_err: float = True,
    ΔmF_absolute=False,
) -> Union[bool, Tuple[bool, str]]:
    """Check whether the transition is allowed based on the quantum numbers

    Args:
        ground_state (CoupledBasisState): ground CoupledBasisState
        state2 (CoupledBasisState): excited CoupledBasisState
        ΔmF_allowed (int): allowed ΔmF for the transition
        return_err (boolean): boolean flag for returning the error message

    Returns:
        tuple: (allowed boolean, error message)
    """
    assert ground_state.P is not None, "parity is required to be set for ground_state"
    assert excited_state.P is not None, "parity is required to be set for excited_state"

    ΔF = int(excited_state.F - ground_state.F)
    ΔmF = int(excited_state.mF - ground_state.mF)
    if ΔmF_absolute:
        ΔmF = np.abs(ΔmF)
        ΔmF_allowed = np.abs(ΔmF_allowed)
    ΔP = int(excited_state.P - ground_state.P)

    flag_ΔP = abs(ΔP) != 2
    flag_ΔF = abs(ΔF) > 1
    flag_ΔmF = ΔmF != ΔmF_allowed
    flag_ΔFΔmF = (not flag_ΔmF) & ((ΔF == 0) & (ΔmF == 0) & (ground_state.mF == 0))

    errors = ""
    if flag_ΔP:
        errors += "parity invalid"
    if flag_ΔF:
        if len(errors) != 0:
            errors += ", "
        errors += f"ΔF invalid -> ΔF = {ΔF}"
    if flag_ΔmF:
        if len(errors) != 0:
            errors += ", "
        errors += f"ΔmF invalid -> ΔmF = {ΔmF}"

    if flag_ΔFΔmF:
        if len(errors) != 0:
            errors += ", "
        errors += "ΔF = 0 & ΔmF = 0 invalid"

    if len(errors) != 0:
        errors = f"transition not allowed; {errors}"

    if return_err:
        return not (flag_ΔP | flag_ΔF | flag_ΔmF | flag_ΔFΔmF), errors
    else:
        return not (flag_ΔP | flag_ΔF | flag_ΔmF | flag_ΔFΔmF)


def assert_transition_coupled_allowed(
    ground_state: states.CoupledBasisState,
    excited_state: states.CoupledBasisState,
    ΔmF_allowed: int,
) -> bool:
    """Check whether the transition is allowed based on the quantum numbers.
    Raises an AssertionError if the transition is not allowed.

    Args:
        ground_state (CoupledBasisState): ground CoupledBasisState
        excited_state (CoupledBasisState): excited CoupledBasisState

    Returns:
        tuple: allowed boolean
    """
    ret = check_transition_coupled_allowed_polarization(
        ground_state, excited_state, ΔmF_allowed, return_err=True
    )
    if isinstance(ret, tuple):
        allowed, errors = ret
        assert allowed, errors
        return allowed
    else:
        return ret


def ΔmF_allowed(
    polarization: npt.NDArray[np.complex_],
) -> int:
    """
    Generate a tuple of all allowed ΔmF

    Returns:
        int: allowed ΔmF
    """
    # TODO: update to more sophisticated method, not a bunch of ifs
    # TODO: return multiple allowed ΔmF
    if np.all(polarization == polarization_X.vector):
        ΔmF = +1
    elif np.all(polarization == polarization_Y.vector):
        ΔmF = +1
    elif np.all(polarization == polarization_Z.vector):
        ΔmF = 0
    elif np.all(polarization == polarization_σm.vector):
        ΔmF = -1
    elif np.all(polarization == polarization_σp.vector):
        ΔmF = +1
    else:
        raise ValueError(
            f"polarization {polarization} cannot be used for select_main_states, not"
            " yet implemented"
        )
    return ΔmF


def select_main_states(
    ground_states: Sequence[states.CoupledState],
    excited_states: Sequence[states.CoupledState],
    polarization: npt.NDArray[np.complex_],
) -> Tuple[states.CoupledState, states.CoupledState]:
    """Select main states for calculating the transition strength to normalize
    the Rabi rate with

    Args:
        ground_states (Sequence[states.State]): Sequence of ground states for the
                                                transition
        excited_states (Sequence[states.State]): Sequence of excited states for the
                                                transition
        polarization (npt.NDArray[np.float_]): polarization vector
    """
    ΔmF = ΔmF_allowed(polarization)

    allowed_transitions = []
    indices_gnd_mF0 = []
    for ide, exc in enumerate(excited_states):
        exc_basisstate = cast(CoupledBasisState, exc.largest)
        for idg, gnd in enumerate(ground_states):
            gnd_basisstate = cast(CoupledBasisState, gnd.largest)
            if check_transition_coupled_allowed_polarization(
                gnd_basisstate, exc_basisstate, ΔmF, return_err=False
            ):
                allowed_transitions.append((idg, ide, exc_basisstate.mF))
                if gnd_basisstate.mF == 0:
                    indices_gnd_mF0.append((idg, ide, gnd_basisstate.mF))

    assert (
        len(allowed_transitions) > 0
    ), "none of the supplied ground and excited states have allowed transitions"

    if len(indices_gnd_mF0) > 0:
        excited_state = excited_states[indices_gnd_mF0[-1][1]]
        ground_state = ground_states[indices_gnd_mF0[-1][0]]
    else:
        idt = len(allowed_transitions) // 2
        excited_state = excited_states[allowed_transitions[idt][1]]
        ground_state = ground_states[allowed_transitions[idt][0]]

    return ground_state, excited_state



from typing import Sequence, Union

import numpy as np
import numpy.typing as npt
import sympy as smp

from centrex_tlf import hamiltonian, states, transitions

__all__ = ["compact_symbolic_hamiltonian_indices", "generate_qn_compact"]


def compact_symbolic_hamiltonian_indices(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    indices_compact: npt.NDArray[np.int_],
) -> smp.matrices.dense.MutableDenseMatrix:
    """compact a sympy hamiltonian by combining all indices in indices_compact
    into a single state

    Args:
        hamiltonian (sympy matrix): hamiltonian
        indices_compact (list, array): indices to compact

    Returns:
        sympy matrix: compacted hamiltonian
    """
    arr = hamiltonian.copy()
    diagonal = arr.diagonal()
    diagonal = [diagonal[idd] for idd in indices_compact]
    # free_symbols = np.unique([val.free_symbols for val in diagonal])
    check_free_symbols = np.sum([len(val.free_symbols) for val in diagonal])
    assert (
        check_free_symbols == 0
    ), "diagonal elements for states to compact have symbols, cannot compact"

    # delete the rows and columns to compact, except a single one that's needed
    # to put the decays into
    deleted = 0
    for idx in indices_compact[1:]:
        row = arr[idx - deleted, :]
        col = arr[:, idx - deleted]
        # check if couplings are present, raise AssertionError if true
        assert (
            np.sum(row) - row[idx - deleted]
        ) == 0, "couplings exist for states to compact, cannot compact"
        assert (
            np.sum(col) - row[idx - deleted]
        ) == 0, "couplings exist for states to compact, cannot compact"
        arr.row_del(idx - deleted)
        arr.col_del(idx - deleted)
        deleted += 1

    # setting the diagonal element to be the mean off the entire state
    # pretty much irrelevant since the states only have decays and should be
    # far enough away from the others
    arr[idx - deleted, idx - deleted] = np.mean(diagonal)
    return arr


def generate_qn_compact(
    transitions: Sequence[
        Union[transitions.OpticalTransition, transitions.MicrowaveTransition]
    ],
    H_reduced: hamiltonian.reduced_hamiltonian.ReducedHamiltonianTotal,
):
    J_transitions_ground = []
    for transition in transitions:
        J_transitions_ground.append(transition.J_ground)
    J_compact = [
        Ji
        for Ji in np.unique([s.J for s in H_reduced.X_states_basis])
        if Ji not in J_transitions_ground
    ]
    qn_compact = [
        states.QuantumSelector(J=Ji, electronic=states.ElectronicState.X)
        for Ji in J_compact
    ]

    return qn_compact



import warnings
from dataclasses import dataclass
from typing import Callable, List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import numpy.typing as npt
from scipy import linalg

from centrex_tlf.states import (
    Basis,
    CoupledBasisState,
    CoupledState,
    ElectronicState,
    QuantumSelector,
    TlFNuclearSpins,
    find_exact_states,
    generate_uncoupled_states_ground,
    get_unique_basisstates_from_states,
)
from centrex_tlf.states.generate_states import (
    generate_coupled_states_B,
    generate_coupled_states_ground,
    generate_coupled_states_X,
)
from centrex_tlf.transitions import MicrowaveTransition, OpticalTransition

from .basis_transformations import generate_transform_matrix
from .constants import BConstants, XConstants
from .generate_hamiltonian import (
    generate_coupled_hamiltonian_B,
    generate_coupled_hamiltonian_B_function,
    generate_uncoupled_hamiltonian_X,
    generate_uncoupled_hamiltonian_X_function,
)
from .matrix_elements import generate_ED_ME_mixed_state
from .utils import matrix_to_states, reduced_basis_hamiltonian, reorder_evecs

__all__ = [
    "generate_diagonalized_hamiltonian",
    "generate_reduced_X_hamiltonian",
    "generate_reduced_B_hamiltonian",
    "compose_reduced_hamiltonian",
    "generate_total_reduced_hamiltonian",
    "generate_reduced_hamiltonian_transitions",
]


@dataclass
class HamiltonianDiagonalized:
    H: npt.NDArray[np.complex128]
    V: npt.NDArray[np.complex128]
    V_ref: Optional[npt.NDArray[np.complex128]] = None


def generate_diagonalized_hamiltonian(
    hamiltonian: npt.NDArray[np.complex128],
    keep_order: bool = True,
    return_V_ref: bool = False,
    rtol: Optional[float] = None,
) -> HamiltonianDiagonalized:
    _ = np.linalg.eigh(hamiltonian)
    D: npt.NDArray[np.complex128] = _[0]
    V: npt.NDArray[np.complex128] = _[1]
    if keep_order:
        V_ref = np.eye(V.shape[0], dtype=np.complex128)
        D, V = reorder_evecs(V, D, V_ref)

    hamiltonian_diagonalized = V.conj().T @ hamiltonian @ V
    if rtol:
        hamiltonian_diagonalized[
            np.abs(hamiltonian_diagonalized)
            < np.abs(hamiltonian_diagonalized).max() * rtol
        ] = 0
    if not return_V_ref or not keep_order:
        return HamiltonianDiagonalized(hamiltonian_diagonalized, V)
    else:
        return HamiltonianDiagonalized(hamiltonian_diagonalized, V, V_ref)


@dataclass
class ReducedHamiltonian:
    H: npt.NDArray[np.complex128]
    V: npt.NDArray[np.complex128]
    QN_basis: List[CoupledState]
    QN_construct: List[CoupledBasisState]

    def __iter__(self):
        # suppot for legacy code
        return iter((self.QN_basis, self.H))


def generate_reduced_X_hamiltonian(
    X_states_approx: Sequence[CoupledBasisState],
    E: npt.NDArray[np.float_] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.float_] = np.array([0.0, 0.0, 1e-3]),
    rtol: Optional[float] = None,
    stol: float = 1e-3,
    Jmin: Optional[int] = None,
    Jmax: Optional[int] = None,
    constants: XConstants = XConstants(),
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    transform: Optional[npt.NDArray[np.complex_]] = None,
    H_func: Optional[Callable] = None,
) -> ReducedHamiltonian:
    """
    Generate the reduced X state hamiltonian.
    Generates the Hamiltonian for all X states from Jmin to Jmax, if provided. Otherwise
    uses the min/max J found in X_states_approx. Then selects the part of the
    hamiltonian that corresponds to X_states_approx.
    The Hamiltonian is diagonal and the returned states are the states corresponding to
    X_states_approx in the basis of the Hamiltonian.

    Args:
        ground_states_approx (Sequence[CoupledBasisState]): States
        E (npt.NDArray[np.float64], optional): Electric field in V/cm. Defaults to
                                                            np.array([0.0, 0.0, 0.0]).
        B (npt.NDArray[np.float64], optional): Magnetic field in G. Defaults to
                                                np.array([0.0, 0.0, 1e-3]). If smaller
                                                than 1e-5 some states become degenerate
                                                again.
        rtol (Optional[float], optional): Remove components smaller than rtol in the
                                                        hamiltonian. Defaults to None.
        stol: (float): Remove superpositions with amplitude smaller than stol from each
                        state. Defaults to 1e-3.
        Jmin (Optional[int], optional): Minimum J to include in the Hamiltonian.
                                        Defaults to None.
        Jmax (Optional[int], optional): Maximum J to include in the Hamiltonian.
                                        Defaults to None.
        constants (XConstants, optional): X state constants. Defaults to XConstants().
        nuclear_spins (TlFNuclearSpins, optional): TlF nuclear spins. Defaults to
                                                                    TlFNuclearSpins().
        transform (npt.NDArray[np.complex_], optional): Transformation matrix from
                                                        uncoupled to coupled for J
                                                        states from Jmin to Jmax.
                                                        Defaults to None.
        H_func (Optional[Callable], optional): Function to generate the Hamiltonian
                                                depending on E and B. Defaults to None.

    Returns:
        ReducedHamiltonian: States and Hamiltonian
    """

    # need to generate the other states in case of mixing
    _Jmin = min([gs.J for gs in X_states_approx]) if Jmin is None else Jmin
    _Jmax = max([gs.J for gs in X_states_approx]) if Jmax is None else Jmax

    QN = generate_uncoupled_states_ground(
        Js=np.arange(_Jmin, _Jmax + 1), nuclear_spins=nuclear_spins
    )
    QNc = generate_coupled_states_ground(
        Js=np.arange(_Jmin, _Jmax + 1), nuclear_spins=nuclear_spins
    )
    if H_func is None:
        H_X_uc = generate_uncoupled_hamiltonian_X(QN, constants=constants)
        H_X_uc_func = generate_uncoupled_hamiltonian_X_function(H_X_uc)
    else:
        H_X_uc_func = H_func

    if transform is None:
        S_transform = generate_transform_matrix(QN, QNc)
    else:
        assert transform.shape[0] == len(QN), (
            (
                f"shape of transform incorrect; requires {len(QN), len(QN)}, "
                f"not {transform.shape}"
            ),
        )
        S_transform = transform

    H_X = S_transform.conj().T @ H_X_uc_func(E, B) @ S_transform
    if rtol:
        H_X[np.abs(H_X) < np.abs(H_X).max() * rtol] = 0

    # diagonalize the Hamiltonian
    H_diagonalized = generate_diagonalized_hamiltonian(
        H_X, keep_order=True, return_V_ref=True, rtol=rtol
    )

    # new set of quantum numbers:
    QN_diag = matrix_to_states(H_diagonalized.V, list(QNc))

    ground_states = find_exact_states(
        [1 * gs for gs in X_states_approx],
        list(QNc),
        QN_diag,
        V=H_diagonalized.V,
        # V_ref=H_diagonalized.V_ref_,
    )
    ground_states = [gs.remove_small_components(stol) for gs in ground_states]

    H_X_red = reduced_basis_hamiltonian(QN_diag, H_diagonalized.H, ground_states)

    return ReducedHamiltonian(
        H=H_X_red, V=H_diagonalized.V, QN_basis=ground_states, QN_construct=list(QNc)
    )


def generate_reduced_B_hamiltonian(
    B_states_approx: Sequence[CoupledBasisState],
    E: npt.NDArray[np.float64] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.float64] = np.array([0.0, 0.0, 1e-5]),
    rtol: Optional[float] = None,
    stol: float = 1e-3,
    Jmin: Optional[int] = None,
    Jmax: Optional[int] = None,
    constants: BConstants = BConstants(),
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    H_func: Optional[Callable] = None,
) -> ReducedHamiltonian:
    """
    Generate the reduced B state hamiltonian.
    Generates the Hamiltonian for all B states from Jmin to Jmax, if provided. Otherwise
    uses Jmin = 1 and for Jmax the maximum J found in B_states_approx with 2 added.
    Then selects the part of the hamiltonian that corresponds to B_states_approx.
    The Hamiltonian is diagonal and the returned states are the states corresponding to
    B_states_approx in the basis of the Hamiltonian.

    Args:
        B_states_approx (Sequence[CoupledBasisState]): States
        E (npt.NDArray[np.float64], optional): Electric field in V/cm. Defaults to
                                                np.array([0.0, 0.0, 0.0]).
        B (npt.NDArray[np.float64], optional): Magnetic field in G. Defaults to
                                                np.array([0.0, 0.0, 1e-5]).
        rtol (Optional[float], optional): Remove components smaller than rtol in the
                                                        hamiltonian. Defaults to None.
        stol: (float): Remove superpositions with amplitude smaller than stol from each
                        state. Defaults to 1e-3.
        Jmin (Optional[int], optional): Minimum J to include in the Hamiltonian.
                                        Defaults to None.
        Jmax (Optional[int], optional): Maximum J to include in the Hamiltonian.
                                        Defaults to None.
        constants (BConstants, optional): B state constants. Defaults to BConstants().
        nuclear_spins (TlFNuclearSpins, optional): TlF nuclear spins. Defaults to
                                                    TlFNuclearSpins().
        H_func (Optional[Callable], optional): Function to generate the Hamiltonian
                                                depending on E and B. Defaults to None.

    Returns:
        ReducedHamiltonian: States and Hamiltonian
    """
    # need to generate the other states in case of mixing
    _Jmin = 1 if Jmin is None else Jmin
    _Jmax = max([gs.J for gs in B_states_approx]) + 2 if Jmax is None else Jmax

    omega_basis_flag = False

    # check basis
    if all([qn.basis == Basis.CoupledP for qn in B_states_approx]):
        qn_select = QuantumSelector(J=np.arange(_Jmin, _Jmax + 1), P=[-1, 1], Ω=1)
    elif all([qn.basis == Basis.CoupledΩ for qn in B_states_approx]):
        qn_select = QuantumSelector(J=np.arange(_Jmin, _Jmax + 1), P=None, Ω=[-1, 1])
        omega_basis_flag = True
    else:
        raise TypeError(
            "B_states_approx basis invalid, CoupledP or CoupledΩ are allowed, not "
            f"{B_states_approx[0].basis}"
        )
    QN_B = list(generate_coupled_states_B(qn_select, nuclear_spins=nuclear_spins))

    if H_func is None:
        H_B = generate_coupled_hamiltonian_B(QN_B, constants=constants)
        H_B_func = generate_coupled_hamiltonian_B_function(H_B)
    else:
        H_B_func = H_func

    H_diagonalized = generate_diagonalized_hamiltonian(
        H_B_func(E, B), keep_order=True, return_V_ref=True, rtol=rtol
    )

    # new set of quantum numbers:
    QN_B_diag = matrix_to_states(H_diagonalized.V, QN_B)

    if omega_basis_flag:
        warnings.warn(
            (
                "generate_reduced_B_hamiltonian called in Ω basis; mapping states to "
                "approximate states not implemented. Hamiltonian is not reduced."
            ),
            SyntaxWarning,
        )
        excited_states = [es.remove_small_components(stol) for es in QN_B_diag]
    else:
        excited_states = find_exact_states(
            [1 * e for e in B_states_approx], QN_B, QN_B_diag, V=H_diagonalized.V
        )
        excited_states = [es.remove_small_components(stol) for es in excited_states]

    H_B_red = reduced_basis_hamiltonian(QN_B_diag, H_diagonalized.H, excited_states)

    return ReducedHamiltonian(
        H=H_B_red,
        V=H_diagonalized.V,
        QN_basis=excited_states,
        QN_construct=list(QN_B),
    )


def compose_reduced_hamiltonian(
    H_X_red: npt.NDArray[np.complex128],
    H_B_red: npt.NDArray[np.complex128],
    element_limit: float = 0.1,
) -> Tuple[npt.NDArray[np.complex128], npt.NDArray[np.complex128]]:
    H_X_red[np.abs(H_X_red) < element_limit] = 0
    H_B_red[np.abs(H_B_red) < element_limit] = 0

    H_int: npt.NDArray[np.complex128] = linalg.block_diag(H_X_red, H_B_red)
    V_ref_int = np.eye(H_int.shape[0], dtype=np.complex128)

    return H_int, V_ref_int


@dataclass
class ReducedHamiltonianTotal:
    X_states: List[CoupledState]
    B_states: List[CoupledState]
    QN: List[CoupledState]
    H_int: npt.NDArray[np.complex_]
    V_ref_int: npt.NDArray[np.complex_]
    X_states_basis: List[CoupledBasisState]
    B_states_basis: List[CoupledBasisState]
    QN_basis: List[CoupledBasisState]


def generate_total_reduced_hamiltonian(
    X_states_approx: Sequence[CoupledBasisState],
    B_states_approx: Sequence[CoupledBasisState],
    E: npt.NDArray[np.float_] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.float_] = np.array([0.0, 0.0, 1e-5]),
    rtol: Optional[float] = None,
    stol: float = 1e-3,
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    X_constants: XConstants = XConstants(),
    B_constants: BConstants = BConstants(),
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    transform: Optional[npt.NDArray[np.complex_]] = None,
    H_func_X: Optional[Callable] = None,
    H_func_B: Optional[Callable] = None,
    use_omega_basis: bool = True,
) -> ReducedHamiltonianTotal:
    """
    Generate the total reduced hamiltonian for all X and B states in X_states_approx and
    B_states_approx, from an X state hamiltonian for all states from Jmin_X to Jmax_X
    and a B state Hamiltonian for all states from Jmin_B to Jmax_B.
    Returns the X_states, B_states, total states, Hamiltonian and V_reference_int which
    keeps track of the state ordering.

    Args:
        X_states_approx (Sequence[CoupledBasisState]): X_states_approx to generate the

        B_states_approx (Sequence[CoupledBasisState]): _description_
        E (npt.NDArray[np.float64], optional): Electric field. Defaults to
                                                np.array([0.0, 0.0, 0.0]).
        B (npt.NDArray[np.float64], optional): Magnetic field. Defaults to
                                                np.array([0.0, 0.0, 1e-5]). If smaller
                                                than 1e-5 some X states become
                                                degenerate again.
        rtol (Optional[float], optional): Tolerance for the Hamiltonian. Defaults to
                                            None.
        stol: (float): Remove superpositions with amplitude smaller than stol from each
                        state. Defaults to 1e-3.
        Jmin_X (Optional[int], optional): Jmin for the X state Hamiltonian. Defaults to
                                            None.
        Jmax_X (Optional[int], optional): Jmax for the X state Hamiltonian. Defaults to
                                            None.
        Jmin_B (Optional[int], optional): Jmin for the B state Hamiltonian. Defaults to
                                            None.
        Jmax_B (Optional[int], optional): Jmax for the B state Hamiltonian. Defaults to
                                            None.
        X_constants (XConstants, optional): X state constants. Defaults to XConstants().
        B_constants (BConstants, optional): B state constants. Defaults to BConstants().
        nuclear_spins (TlFNuclearSpins, optional): TlF nuclear spins. Defaults to
                                                    TlFNuclearSpins().
        transform (npt.NDArray[np.complex128], optional): transformation matrix to
                                                            transform the uncoupled X
                                                            state Hamiltonian to
                                                            coupled. Defaults to None.
        H_func_X (Optional[Callable], optional): Function to generate the X state
                                                    Hamiltonian for E and B. Defaults
                                                    to None.
        H_func_B (Optional[Callable], optional): Function to generate the B state
                                                    Hamiltonian for E and B. Defaults
                                                    to None.
        use_omega_basis (Optional[bool]): Use Ω basis to calculate the B state
                                            hamiltonian and then transform to P basis if
                                            the P state is requested. Ω basis is faster
                                            to calculate. Defaults to True.

    Returns:
        ReducedHamiltonian: Dataclass holding the X states, B states, total states,
                            Hamiltonian and reference eigenvectors
    """

    ground_states, H_X_red = generate_reduced_X_hamiltonian(
        X_states_approx,
        E=E,
        B=B,
        rtol=rtol,
        stol=stol,
        Jmin=Jmin_X,
        Jmax=Jmax_X,
        constants=X_constants,
        nuclear_spins=nuclear_spins,
        H_func=H_func_X,
        transform=transform,
    )

    if use_omega_basis and B_states_approx[0].basis == Basis.CoupledP:
        _B_states_approx = cast(
            Sequence[CoupledBasisState],
            get_unique_basisstates_from_states(
                [qn.transform_to_omega_basis() for qn in B_states_approx]
            ),
        )
    else:
        _B_states_approx = cast(List[CoupledBasisState], list(B_states_approx))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        H_B_red = generate_reduced_B_hamiltonian(
            _B_states_approx,
            E=E,
            B=B,
            rtol=rtol,
            stol=stol,
            Jmin=Jmin_B,
            Jmax=Jmax_B,
            constants=B_constants,
            nuclear_spins=nuclear_spins,
            H_func=H_func_B,
        )

    if use_omega_basis and B_states_approx[0].basis == Basis.CoupledP:
        excited_states = [
            qn.transform_to_parity_basis().remove_small_components(stol)
            for qn in H_B_red.QN_basis
        ]

        state_vecs = np.array(
            [(1 * s).state_vector(excited_states) for s in B_states_approx]
        )
        excited_states_reduced = [
            excited_states[idx] for idx in np.argmax(np.abs(state_vecs), axis=1)
        ]

        H_B_red_matrix = reduced_basis_hamiltonian(
            excited_states, H_B_red.H, excited_states_reduced
        )

        excited_states = excited_states_reduced

        H_int, V_ref_int = compose_reduced_hamiltonian(H_X_red, H_B_red_matrix)
    else:
        excited_states = H_B_red.QN_basis
        H_int, V_ref_int = compose_reduced_hamiltonian(H_X_red, H_B_red.H)

    QN = ground_states.copy()
    QN.extend(excited_states)
    return ReducedHamiltonianTotal(
        X_states=ground_states,
        B_states=excited_states,
        QN=QN,
        H_int=H_int,
        V_ref_int=V_ref_int,
        X_states_basis=list(X_states_approx),
        B_states_basis=list(B_states_approx),
        QN_basis=list(X_states_approx) + list(B_states_approx),
    )


def generate_reduced_hamiltonian_transitions(
    transitions: Sequence[Union[OpticalTransition, MicrowaveTransition]],
    E: npt.NDArray[np.float64] = np.array([0.0, 0.0, 0.0]),
    B: npt.NDArray[np.float64] = np.array([0.0, 0.0, 1e-5]),
    rtol: Optional[float] = None,
    stol: float = 1e-3,
    Jmin_X: Optional[int] = None,
    Jmax_X: Optional[int] = None,
    Jmin_B: Optional[int] = None,
    Jmax_B: Optional[int] = None,
    Xconstants: XConstants = XConstants(),
    Bconstants: BConstants = BConstants(),
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    minimum_amplitude: float = 5e-3,
    minimum_coupling: float = 1e-3,
    use_omega_basis: bool = True,
) -> ReducedHamiltonianTotal:
    _J_ground: List[int] = []
    excited_states_selectors = []

    # figure out which rotational levels to include
    for transition in transitions:
        if isinstance(transition, OpticalTransition):
            excited_states_approx_qn_select = transition.qn_select_excited
            excited_states_approx = list(
                generate_coupled_states_B(excited_states_approx_qn_select)
            )
            excited_states, excited_hamiltonian = generate_reduced_B_hamiltonian(
                B_states_approx=excited_states_approx,
                E=E,
                B=B,
                rtol=rtol,
                stol=stol,
                Jmin=Jmin_B,
                Jmax=Jmax_B,
                constants=Bconstants,
                nuclear_spins=nuclear_spins,
            )

            # figure out which excited states are involved
            excited_states = [
                s.remove_small_components(minimum_amplitude) for s in excited_states
            ]

            Js_excited: npt.NDArray[np.int_] = np.unique(
                [s.J for es in excited_states for a, s in es]
            )
            if Jmax_X is None:
                _Jmax_X = Js_excited.max() + 2
            else:
                _Jmax_X = Jmax_X
            ground_states = generate_coupled_states_ground(
                Js=np.arange(0, _Jmax_X + 1).astype(int)
            )
            # calculate the coupling between the ground and excited states to determine
            # which states to include
            nonzero_coupling = []
            for gs in ground_states:
                for es in excited_states:
                    if (
                        np.abs(generate_ED_ME_mixed_state(1 * gs, es))
                        >= minimum_coupling
                    ):
                        nonzero_coupling.append(gs)

            Js_ground = list(np.unique([s.J for s in nonzero_coupling]))

            _J_ground.extend(Js_ground)
            excited_states_selectors.append(excited_states_approx_qn_select)

        if isinstance(transition, MicrowaveTransition):
            _J_ground.extend([transition.J_ground, transition.J_excited])

    # removing duplicates
    J_ground: List[int] = list(np.unique(_J_ground))

    ground_states_approx_qn_select = QuantumSelector(
        J=J_ground, electronic=ElectronicState.X, Ω=0
    )
    ground_states_approx = list(
        generate_coupled_states_X(ground_states_approx_qn_select)
    )
    excited_states_approx = list(generate_coupled_states_B(excited_states_selectors))

    return generate_total_reduced_hamiltonian(
        X_states_approx=ground_states_approx,
        B_states_approx=excited_states_approx,
        E=E,
        B=B,
        rtol=rtol,
        stol=stol,
        Jmin_X=Jmin_X,
        Jmax_X=Jmax_X,
        Jmin_B=Jmin_B,
        Jmax_B=Jmax_B,
        X_constants=Xconstants,
        B_constants=Bconstants,
        nuclear_spins=nuclear_spins,
        transform=None,
        H_func_X=None,
        H_func_B=None,
        use_omega_basis=use_omega_basis,
    )



from dataclasses import dataclass
from itertools import product
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Sequence,
    TypeVar,
    Union,
    no_type_check,
    overload,
)

import numpy as np
import numpy.typing as npt

from .states import (
    BasisState,
    CoupledBasisState,
    CoupledState,
    ElectronicState,
    UncoupledBasisState,
    UncoupledState,
)
from .utils import get_unique_list, reorder_evecs

__all__ = [
    "QuantumSelector",
    "find_state_idx_from_state",
    "find_exact_states_indices",
    "find_exact_states",
    "find_closest_vector_idx",
    "check_approx_state_exact_state",
    "get_indices_quantumnumbers_base",
    "get_indices_quantumnumbers",
    "get_unique_basisstates_from_basisstates",
    "get_unique_basisstates_from_states",
]


@dataclass
class QuantumSelector:
    """Class for setting quantum numbers for selecting a subset of states
    from a larger set of states

    Args:
        J (Union[NumberType, list, np.ndarray]): rotational quantum number
        F1 (Union[NumberType, list, np.ndarray]):
        F (Union[NumberType, list, np.ndarray]):
        mF (Union[NumberType, list, np.ndarray]):
        electronic (Union[str, list, np.ndarray]): electronic state
    """

    J: Optional[Union[Sequence[int], npt.NDArray[np.int_], int]] = None
    F1: Optional[Union[Sequence[float], npt.NDArray[np.float_], float]] = None
    F: Optional[Union[Sequence[int], npt.NDArray[np.int_], int]] = None
    mF: Optional[Union[Sequence[int], npt.NDArray[np.int_], float]] = None
    electronic: Optional[
        Union[Sequence[ElectronicState], npt.NDArray[Any], ElectronicState]
    ] = None
    P: Optional[Union[Callable, Sequence[int], npt.NDArray[np.int_], int]] = None
    Ω: Optional[Union[Sequence[int], npt.NDArray[np.int_], int]] = None

    def get_indices(
        self,
        QN: Union[
            Sequence[CoupledState], Sequence[CoupledBasisState], npt.NDArray[Any]
        ],
        mode: str = "python",
    ) -> npt.NDArray[np.int_]:
        return get_indices_quantumnumbers_base(self, QN, mode)


def find_state_idx_from_state(
    H: npt.NDArray[np.complex128],
    reference_state: CoupledState,
    QN: Sequence[CoupledState],
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> int:
    """Determine the index of the state vector most closely corresponding to an
    input state

    Args:
        H (np.ndarray): Hamiltonian to compare to
        reference_state (State): state to find closest state in H to
        QN (list): list of state objects defining the basis for H

    Returns:
        int: index of closest state vector of H corresponding to reference_state
    """

    # determine state vector of reference state
    reference_state_vec = reference_state.state_vector(QN)

    # find eigenvectors of the given Hamiltonian
    _ = np.linalg.eigh(H)
    E: npt.NDArray[np.complex128] = _[0]
    V: npt.NDArray[np.complex128] = _[1]

    if V_ref is not None:
        E, V = reorder_evecs(V, E, V_ref)

    overlaps = np.dot(np.conj(reference_state_vec), V)
    probabilities = overlaps * np.conj(overlaps)

    idx = int(np.argmax(probabilities))

    return idx


def find_closest_vector_idx(
    state_vec: npt.NDArray[np.complex128], vector_array: npt.NDArray[np.complex128]
) -> int:
    """Function that finds the index of the vector in vector_array that most closely
    matches state_vec. vector_array is array where each column is a vector, typically
    corresponding to an eigenstate of some Hamiltonian.

    inputs:
    state_vec = Numpy array, 1D
    vector_array = Numpy array, 2D

    returns:
    idx = index that corresponds to closest matching vector
    """

    overlaps = np.abs(state_vec.conj().T @ vector_array)
    idx = int(np.argmax(overlaps))

    return idx


def check_approx_state_exact_state(approx: CoupledState, exact: CoupledState) -> None:
    """Check if the exact found states match the approximate states. The exact
    states are found from the eigenvectors of the hamiltonian and are often a
    superposition of various states.
    The approximate states are used in initial setup of the hamiltonian.

    Args:
        approx (State): approximate state
        exact (State): exact state
    """
    approx_largest = approx.find_largest_component()
    exact_largest = exact.find_largest_component()

    if not type(approx_largest) == type(exact_largest):
        raise TypeError(
            f"can't compare approx ({type(approx_largest)}) and exact "
            f"({type(exact_largest)}), not equal types"
        )

    if isinstance(approx_largest, CoupledBasisState) and isinstance(
        exact_largest, CoupledBasisState
    ):
        assert approx_largest.electronic_state == exact_largest.electronic_state, (
            f"mismatch in electronic state {approx_largest.electronic_state} != "
            f"{exact_largest.electronic_state}"
        )
        assert (
            approx_largest.J == exact_largest.J
        ), f"mismatch in J {approx_largest.J} != {exact_largest.J}"
        assert (
            approx_largest.F == exact_largest.F
        ), f"mismatch in F {approx_largest.F} != {exact_largest.F}"
        assert (
            approx_largest.F1 == exact_largest.F1
        ), f"mismatch in F1 {approx_largest.F1} != {exact_largest.F1}"
        assert (
            approx_largest.mF == exact_largest.mF
        ), f"mismatch in mF {approx_largest.mF} != {exact_largest.mF}"
    else:
        raise NotImplementedError


@overload
def find_exact_states_indices(
    states_approx: CoupledState,
    QN_construct: CoupledBasisState,
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> npt.NDArray[np.int_]: ...


@overload
def find_exact_states_indices(
    states_approx: UncoupledState,
    QN_construct: UncoupledBasisState,
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> npt.NDArray[np.int_]: ...


def find_exact_states_indices(
    states_approx,
    QN_construct,
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> npt.NDArray[np.int_]:
    """
    Find the indices for the closest approximate eigenstates corresponding to
    states_approx for a Hamiltonian constructed from the quantum states QN_approx.

    Args:
        states_approx (Sequence[State]): approximate states to find the indices for
        QN_construct (Sequence[State]): states from which H was constructed
        H (Optional[npt.NDArray[np.complex128]], optional): Hamiltonian. Defaults to
                                                                                None.
        V (Optional[npt.NDArray[np.complex128]], optional): Eigenvectors. Defaults to
                                                                                None.
        V_ref (Optional[npt.NDArray[np.complex128]], optional): Eigenvector order.
                                                                Defaults to None.

    Returns:
        npt.NDArray[np.int_]: indices
    """
    # generating the state vectors for states_approx in the basis of QN, which is the
    # basis H was generated from. Note that this is not the actual basis of H, which
    # can be generated with QN and V, with the matrix_to_states function in .utils
    state_vecs = np.array([s.state_vector(QN_construct) for s in states_approx])

    # generate the eigenvectors if they were not provided
    if V is None:
        assert H is not None, "Need to supply H if V is None"
        _V = np.linalg.eigh(H)[1]
    else:
        _V = V
    if V_ref is not None:
        _, _V = reorder_evecs(_V, np.ones(len(QN_construct), dtype=np.complex_), V_ref)

    # calculating the overlaps between the state vectors of states_approx in the basis
    # from which H was generated with the eigenvectors of H
    overlaps = np.dot(np.conj(state_vecs), _V) ** 2
    indices = np.argmax(overlaps, axis=1)

    # check if all maximal overlaps are unique
    if np.unique(indices).size != len(indices):
        raise AssertionError(f"duplicate indices found: {list(indices)}")
    return indices


@overload
def find_exact_states(
    states_approx: Sequence[CoupledState],
    QN_construct: Union[Sequence[CoupledBasisState], Sequence[CoupledState]],
    QN_basis: Sequence[CoupledState],
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> List[CoupledState]: ...


@overload
def find_exact_states(
    states_approx: Sequence[UncoupledState],
    QN_construct: Union[Sequence[UncoupledBasisState], Sequence[UncoupledState]],
    QN_basis: Sequence[UncoupledState],
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
) -> List[UncoupledState]: ...


def find_exact_states(
    states_approx,
    QN_construct,
    QN_basis,
    H: Optional[npt.NDArray[np.complex128]] = None,
    V: Optional[npt.NDArray[np.complex128]] = None,
    V_ref: Optional[npt.NDArray[np.complex128]] = None,
):
    """Find closest approximate eigenstates corresponding to states_approx

    Args:
        states_approx (list): list of State objects to find the closest match to
        QN_construct (list): list of BasisState objects from which H was constructed
        QN_basis (list): list of State objects defining the basis for H
        H (np.ndarray): Hamiltonian, diagonal in basis QN_basis
        V (np.ndarray): eigenvectors in basis QN_basis

    Returns:
        list: list of eigenstates of H closest to states_approx
    """

    indices = find_exact_states_indices(states_approx, QN_construct, H, V, V_ref)
    return [QN_basis[idx] for idx in indices]


@no_type_check
def get_indices_quantumnumbers_base(
    qn_selector: QuantumSelector,
    QN: Union[Sequence[CoupledState], Sequence[CoupledBasisState], npt.NDArray[Any]],
    mode: str = "python",
) -> npt.NDArray[np.int_]:
    """Return the indices corresponding to all states in QN that correspond to
    the quantum numbers in QuantumSelector.
    Entries in QuantumSelector quantum numbers can be either single numbers or
    lists/arrays. States with all possible combinations of quantum numbers in
    QuantumSelector are found

    Args:
        qn_selector (QuantumSelector): QuantumSelector class containing the
                                        quantum numbers to find corresponding
                                        indices for
        QN (Union[list, np.ndarray]): list or array of states

    Raises:
        AssertionError: only supports State and CoupledBasisState types the States list
        or array

    Returns:
        np.ndarray: indices corresponding to the quantum numbers
    """
    assert isinstance(
        qn_selector, QuantumSelector
    ), "supply a QuantumSelector object to select states"
    if isinstance(QN[0], CoupledState):
        Js = np.array([s.largest.J for s in QN])
        F1s = np.array([s.largest.F1 for s in QN])
        Fs = np.array([s.largest.F for s in QN])
        mFs = np.array([s.largest.mF for s in QN])
        estates = np.array([s.largest.electronic_state for s in QN])
    elif isinstance(QN[0], CoupledBasisState):
        Js = np.array([s.J for s in QN])
        F1s = np.array([s.F1 for s in QN])
        Fs = np.array([s.F for s in QN])
        mFs = np.array([s.mF for s in QN])
        estates = np.array([s.electronic_state for s in QN])
    else:
        raise AssertionError(
            "get_indices_quantumnumbers_base() only supports State and "
            "CoupledBasisState types the States list or array"
        )

    J = qn_selector.J
    F1 = qn_selector.F1
    F = qn_selector.F
    mF = qn_selector.mF
    estate = qn_selector.electronic
    assert estate is not None, "supply the electronic state to select states"

    # generate all combinations
    fields = []
    for par in ["J", "F1", "F", "mF", "electronic"]:
        par = getattr(qn_selector, par)
        fields.append([par] if not isinstance(par, (list, tuple, np.ndarray)) else par)
    combinations = product(*fields)

    mask = np.zeros(len(QN), dtype=bool)
    mask_all = np.ones(len(QN), dtype=bool)
    for J, F1, F, mF, estate in combinations:
        # generate the masks for states in QN where the conditions are met
        mask_J = Js == J if J is not None else mask_all
        mask_F1 = F1s == F1 if F1 is not None else mask_all
        mask_F = Fs == F if F is not None else mask_all
        mask_mF = mFs == mF if mF is not None else mask_all
        mask_es = (
            estates == estate if estate is not None else np.zeros(len(QN), dtype=bool)
        )
        # get the indices of the states in QN to compact
        mask = mask | (mask_J & mask_F1 & mask_F & mask_mF & mask_es)

    if mode == "python":
        return np.where(mask)[0]
    elif mode == "julia":
        return np.where(mask)[0] + 1
    else:
        raise TypeError("mode != python or julia")


def get_indices_quantumnumbers(
    qn_selector: Union[QuantumSelector, Sequence[QuantumSelector], npt.NDArray[Any]],
    QN: Union[Sequence[CoupledState], Sequence[CoupledBasisState], npt.NDArray[Any]],
) -> npt.NDArray[np.int_]:
    """return the indices corresponding to all states in QN that correspond to
    the quantum numbers in QuantumSelector or a list of QuantumSelector objects.
    Entries in QuantumSelector quantum numbers can be either single numbers or
    lists/arrays. States with all possible combinations of quantum numbers in
    QuantumSelector are found

    Args:
        qn_selector (Union[QuantumSelector, list, np.ndarray]):
                    QuantumSelector class or list/array of QuantumSelectors
                    containing the quantum numbers to find corresponding indices

        QN (Union[list, np.ndarray]): list or array of states

    Raises:
        AssertionError: only supports State and CoupledBasisState types the States list
        or array

    Returns:
        np.ndarray: indices corresponding to the quantum numbers
    """
    if isinstance(qn_selector, QuantumSelector):
        return get_indices_quantumnumbers_base(qn_selector, QN)
    elif isinstance(qn_selector, (list, np.ndarray)):
        return np.unique(
            np.concatenate(
                [get_indices_quantumnumbers_base(qns, QN) for qns in qn_selector]
            )
        )
    else:
        raise AssertionError(
            "qn_selector required to be of type QuantumSelector, list or np.ndarray"
        )


StateType = TypeVar("StateType")


def get_unique_basisstates_from_basisstates(
    states: Sequence[StateType],
) -> List[StateType]:
    """get a list/array of unique BasisStates in the states list/array

    Args:
        states (Sequence): list/array of BasisStates

    Returns:
        List: list/array of unique BasisStates
    """
    assert isinstance(states[0], BasisState), "Not a sequence of BasisState objects"
    return get_unique_list(states)


def get_unique_basisstates_from_states(
    states: Sequence[CoupledState],
) -> List[BasisState]:
    """
    get a Sequence of unique BasisStates in a sequence of State objects

    Args:
        states (Sequence[State]): Sequence of State objects

    Returns:
        Sequence[BasisState]: Sequence of unique BasisStates that comprise the input
                                States
    """
    assert isinstance(states[0], CoupledState), "Not a sequence of State objects"
    return get_unique_basisstates_from_basisstates(
        [s for S in states for a, s in S.data]
    )



from dataclasses import dataclass

import numpy as np

__all__ = ["HamiltonianConstants", "XConstants", "BConstants", "Γ"]

Γ = 2 * np.pi * 1.56e6

# Values for rotational constant are from "Microwave Spectral tables:
# Diatomic molecules" by Lovas & Tiemann (1974).
# Note that Brot differs from the one given by Ramsey by about 30 MHz.

B_ϵ = 6.689873e9
α = 45.0843e6


@dataclass
class HamiltonianConstants:
    B_rot: float


@dataclass(unsafe_hash=True)
class XConstants(HamiltonianConstants):
    B_rot: float = B_ϵ - α / 2
    c1: float = 126030.0
    c2: float = 17890.0
    c3: float = 700.0
    c4: float = -13300.0
    μ_J: float = 35.0  # Hz/G
    μ_Tl: float = 1240.5  # Hz/G
    μ_F: float = 2003.63  # Hz/G
    D_TlF: float = 4.2282 * 503411.7791722602  # Convert Debye to Hz/(V/cm)


@dataclass(unsafe_hash=True)
class BConstants(HamiltonianConstants):
    # Constants in Hz
    B_rot: float = 6687.879e6
    D_rot: float = 0.010869e6
    H_const: float = -8.1e-2
    h1_Tl: float = 28789e6
    h1_F: float = 861e6
    q: float = 2.423e6
    c_Tl: float = -7.83e6
    c1p_Tl: float = 11.17e6
    μ_B: float = 1.4e6
    gL: float = 1
    gS: float = 2
    μ_E = 2.28 * 503411.7791722602  # Convert Debye to Hz/(V/cm)



import copy
from typing import Any, List, Literal, Optional, Sequence, Tuple, Union, overload

import numpy as np
import numpy.typing as npt
import sympy as smp

from centrex_tlf import couplings as couplings_tlf
from centrex_tlf import states

from .utils_compact import compact_symbolic_hamiltonian_indices

__all__ = [
    "symbolic_hamiltonian_to_rotating_frame",
    "generate_symbolic_hamiltonian",
    "generate_total_symbolic_hamiltonian",
]


def symbolic_hamiltonian_to_rotating_frame(
    hamiltonian: smp.matrices.dense.MutableDenseMatrix,
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: Sequence[couplings_tlf.CouplingFields],
    δs: Sequence[smp.Symbol],
) -> smp.matrices.dense.MutableDenseMatrix:
    """Transform a symbolic hamiltonian to the rotating frame. Exponential terms
    with the transition frequencies are required to be present in the
    hamiltonian matrix, as well as symbolic energies on the diagonal.

    Args:
        hamiltonian (sympy.Matrix): symbolic hamiltonian
        QN (list/array): list/array of states in the system
        H_int (np.ndarray): numerical hamiltonian, energies only
        couplings (list): list of couplings in system

    Returns:
        sympy.Matrix: symbolic hamiltonian in the rotating frame
    """
    n_states = H_int.shape[0]
    energies = np.diag(hamiltonian)

    # generate t symbol for non-rotating frame
    t = smp.Symbol("t", real=True)

    coupled_states = []
    for i, j in zip(*np.nonzero(hamiltonian)):
        if i < j:
            syms = hamiltonian[i, j].free_symbols
            syms = [s for s in syms if str(s)[0] == "ω"]
            assert len(syms) == 1, f"Too many/few couplings, syms = {syms}"
            coupled_states.append((i, j, syms[0]))

    # solve equations to generate unitary transformation to rotating frame
    A = smp.symbols(f"a:{n_states}")
    Eqns = []
    # generate equations
    for i, j, ω in coupled_states:
        Eqns.append(ω - (A[i] - A[j]))
    # solve system of equations
    sol = smp.solve(Eqns, A)
    # set free parameters to zero in the solution
    free_params = [value for value in A if value not in list(sol.keys())]
    for free_param in free_params:
        for key, val in sol.items():
            sol[key] = val.subs(free_param, 0)

    # generate unitary transformation matrix
    T = smp.eye(*H_int.shape)
    for var in sol.keys():
        ida = int(str(var)[1:])
        T[ida, ida] = smp.exp(1j * sol[var] * t)

    # use unitary matrix to transform to rotating frame
    transformed = T.adjoint() @ hamiltonian @ T - 1j * T.adjoint() @ smp.diff(T, t)
    transformed = smp.simplify(transformed)

    transformed = smp.Matrix(transformed)

    for idc, (δ, coupling) in enumerate(zip(δs, couplings)):
        # generate transition frequency symbol
        ω = smp.Symbol(f"ω{idc}", real=True)
        # get indices of ground and excited states
        idg = QN.index(coupling.ground_main)
        ide = QN.index(coupling.excited_main)

        # transform to δ instead of ω and E
        if idg < ide:
            transformed = transformed.subs(ω, energies[ide] - energies[idg] + δ)
        elif idg > ide:
            transformed = transformed.subs(ω, energies[idg] - energies[ide] + δ)

    # remove excited state energy from all diagonal entries
    for idc, (δ, coupling) in enumerate(zip(δs, couplings)):
        idg = QN.index(coupling.ground_main)
        expr = transformed[idg, idg]
        for d in δs:
            expr = expr.subs(d, 0)

        for idx in range(transformed.shape[0]):
            transformed[idx, idx] -= expr

    # remove extraneous 1.0 from expressions like 1.0*δ
    for idx in range(transformed.shape[0]):
        transformed[idx, idx] = smp.nfloat(smp.nsimplify(transformed[idx, idx]))

    # substitute level energies for symbolic values
    transformed = transformed.subs(
        [(E, val) for E, val in zip(energies, np.diag(H_int))]
    )

    return transformed


def generate_symbolic_hamiltonian(
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: Sequence[couplings_tlf.CouplingFields],
    Ωs: Sequence[smp.Symbol],
    δs: Sequence[smp.Symbol],
    pols: List[Optional[Sequence[smp.Symbol]]],
) -> smp.matrices.dense.MutableDenseMatrix:
    n_states = H_int.shape[0]
    # initialize empty hamiltonian
    hamiltonian = smp.zeros(*H_int.shape)
    energies = smp.symbols(f"E:{n_states}")
    hamiltonian += smp.eye(n_states) * np.asarray(energies)

    # generate t symbol for non-rotating frame
    t = smp.Symbol("t", real=True)

    # iterate over couplings
    for idc, (Ω, coupling) in enumerate(zip(Ωs, couplings)):
        # generate transition frequency symbol
        ω = smp.Symbol(f"ω{idc}", real=True)
        # main coupling matrix element
        main_coupling = coupling.main_coupling
        # iterate over fields (polarizations) in the coupling
        for idf, field in enumerate(coupling.fields):
            if pols:
                P = pols[idc]
                if P:
                    _P = P[idf]
                    val = (_P * Ω / main_coupling) / 2
                    for i, j in zip(*np.nonzero(field.field)):
                        if i < j:
                            hamiltonian[i, j] += (
                                val * field.field[i, j] * smp.exp(1j * ω * t)
                            )
                            hamiltonian[j, i] += (
                                val * field.field[j, i] * smp.exp(-1j * ω * t)
                            )
                else:
                    val = (Ω / main_coupling) / 2
                    for i, j in zip(*np.nonzero(field.field)):
                        if i < j:
                            hamiltonian[i, j] += (
                                val * field.field[i, j] * smp.exp(1j * ω * t)
                            )
                            hamiltonian[j, i] += (
                                val * field.field[j, i] * smp.exp(-1j * ω * t)
                            )
            else:
                val = (Ω / main_coupling) / 2
                for i, j in zip(*np.nonzero(field.field)):
                    if i < j:
                        hamiltonian[i, j] += (
                            val * field.field[i, j] * smp.exp(1j * ω * t)
                        )
                        hamiltonian[j, i] += (
                            val * field.field[j, i] * smp.exp(-1j * ω * t)
                        )

    hamiltonian = smp.simplify(hamiltonian)

    transformed = symbolic_hamiltonian_to_rotating_frame(
        hamiltonian, QN, H_int, couplings, δs
    )
    transformed = smp.Matrix(transformed)

    Ωsᶜ = [smp.Symbol(str(Ω) + "ᶜ", complex=True) for Ω in Ωs]
    for idx in range(n_states):
        for idy in range(0, idx):
            for Ω, Ωᶜ in zip(Ωs, Ωsᶜ):
                transformed[idx, idy] = transformed[idx, idy].subs(Ω, Ωᶜ)

    return transformed


@overload
def generate_total_symbolic_hamiltonian(
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: List[couplings_tlf.CouplingFields],
    transitions: Sequence[Any],
    qn_compact: Literal[None],
) -> smp.matrices.dense.MutableDenseMatrix: ...


@overload
def generate_total_symbolic_hamiltonian(
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: List[couplings_tlf.CouplingFields],
    transitions: Sequence[Any],
) -> smp.matrices.dense.MutableDenseMatrix: ...


@overload
def generate_total_symbolic_hamiltonian(
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: Sequence[couplings_tlf.CouplingFields],
    transitions: Sequence[Any],
    qn_compact: Union[Sequence[states.QuantumSelector], states.QuantumSelector],
) -> Tuple[smp.matrices.dense.MutableDenseMatrix, List[states.CoupledState]]: ...


def generate_total_symbolic_hamiltonian(
    QN: List[states.CoupledState],
    H_int: npt.NDArray[np.complex_],
    couplings: Sequence[Any],
    transitions: Sequence[Any],
    qn_compact: Optional[
        Union[Sequence[states.QuantumSelector], states.QuantumSelector]
    ] = None,
) -> Union[
    Tuple[smp.matrices.dense.MutableDenseMatrix, List[states.CoupledState]],
    smp.matrices.dense.MutableDenseMatrix,
]:
    """Generate the total symbolic hamiltonian for the given system

    Args:
        QN (Sequence[states.State]): states
        H_int (np.ndarray): internal hamiltonian
        couplings (Sequence[states.State]): list of dictionaries with all couplings of
                                            the system
        transitions (Sequence[states.State]): list of dictionaries with all transitions
                                                of the system
        qn_compact (Sequence[states.State], optional): list of QuantumSelectors or lists
                                                        of QuantumSelectors with each
                                                        QuantumSelector containing the
                                                        quantum numbers to compact into
                                                        a single state.
                                                        Defaults to None.

    Returns:
        sympy matrix: symbolic hamiltonian
        if qn_compact is provided, also returns the states corresponding to the
        compacted hamiltonian, i.e. ham, QN_compact
    """
    Ωs = [t.Ω for t in transitions]
    Δs = [t.δ for t in transitions]
    pols: List[Optional[Sequence[smp.Symbol]]] = []
    for transition in transitions:
        if not transition.polarization_symbols:
            pols.append(None)
        else:
            pols.append(transition.polarization_symbols)

    H_symbolic = generate_symbolic_hamiltonian(QN, H_int, couplings, Ωs, Δs, pols)
    if qn_compact is not None:
        if isinstance(qn_compact, states.QuantumSelector):
            qn_compact = [qn_compact]
        QN_compact = copy.deepcopy(QN)
        for qnc in qn_compact:
            indices_compact = states.get_indices_quantumnumbers(qnc, QN_compact)
            QN_compact = states.compact_QN_coupled_indices(QN_compact, indices_compact)
            H_symbolic = compact_symbolic_hamiltonian_indices(
                H_symbolic, indices_compact
            )
        return H_symbolic, QN_compact

    return H_symbolic



import math
from functools import lru_cache
from typing import Callable, overload

from centrex_tlf.states import (
    CoupledBasisState,
    CoupledState,
    UncoupledBasisState,
    UncoupledState,
)

from .constants import HamiltonianConstants

__all__ = [
    "J2",
    "J4",
    "J6",
    "I1z",
    "I2z",
    "Jp",
    "Jm",
    "I1p",
    "I1m",
    "I2p",
    "I2m",
    "Jx",
    "Jy",
    "I1x",
    "I1y",
    "I2x",
    "I2y",
    "com",
]

########################################################
# Diagonal operators multiple state by eigenvalue
########################################################


@overload
def J2(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def J2(psi: CoupledBasisState, *args) -> CoupledState: ...


def J2(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([(psi.J * (psi.J + 1), psi)])


@overload
def J4(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def J4(psi: CoupledBasisState, *args) -> CoupledState: ...


def J4(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([((psi.J * (psi.J + 1)) ** 2, psi)])


@overload
def J6(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def J6(psi: CoupledBasisState, *args) -> CoupledState: ...


def J6(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([((psi.J * (psi.J + 1)) ** 3, psi)])


@overload
def Jz(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def Jz(psi: CoupledBasisState, *args) -> CoupledState: ...


def Jz(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([(psi.mJ, psi)])


@overload
def I1z(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def I1z(psi: CoupledBasisState, *args) -> CoupledState: ...


def I1z(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([(psi.m1, psi)])


@overload
def I2z(psi: UncoupledBasisState, *args) -> UncoupledState: ...


@overload
def I2z(psi: CoupledBasisState, *args) -> CoupledState: ...


def I2z(psi, *args):
    if isinstance(psi, CoupledBasisState):
        psi_class = CoupledState
    else:
        psi_class = UncoupledState
    return psi_class([(psi.m2, psi)])


########################################################
#
########################################################


def Jp(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.J - psi.mJ) * (psi.J + psi.mJ + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ + 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


def Jm(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.J + psi.mJ) * (psi.J - psi.mJ + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ - 1,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


def I1p(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.I1 - psi.m1) * (psi.I1 + psi.m1 + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ,
        psi.I1,
        psi.m1 + 1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


def I1m(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.I1 + psi.m1) * (psi.I1 - psi.m1 + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ,
        psi.I1,
        psi.m1 - 1,
        psi.I2,
        psi.m2,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


def I2p(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.I2 - psi.m2) * (psi.I2 + psi.m2 + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2 + 1,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


def I2m(psi: UncoupledBasisState, *args) -> UncoupledState:
    amp = math.sqrt((psi.I2 + psi.m2) * (psi.I2 - psi.m2 + 1))
    ket = UncoupledBasisState(
        psi.J,
        psi.mJ,
        psi.I1,
        psi.m1,
        psi.I2,
        psi.m2 - 1,
        Omega=psi.Omega,
        P=psi.P,
        electronic_state=psi.electronic_state,
    )
    return UncoupledState([(amp, ket)])


########################################################
###
########################################################


def Jx(psi: UncoupledBasisState, *args) -> UncoupledState:
    return 0.5 * (Jp(psi) + Jm(psi))


def Jy(psi: UncoupledBasisState, *args) -> UncoupledState:
    return -0.5j * (Jp(psi) - Jm(psi))


def I1x(psi: UncoupledBasisState, *args) -> UncoupledState:
    return 0.5 * (I1p(psi) + I1m(psi))


def I1y(psi: UncoupledBasisState, *args) -> UncoupledState:
    return -0.5j * (I1p(psi) - I1m(psi))


def I2x(psi: UncoupledBasisState, *args) -> UncoupledState:
    return 0.5 * (I2p(psi) + I2m(psi))


def I2y(psi: UncoupledBasisState, *args) -> UncoupledState:
    return -0.5j * (I2p(psi) - I2m(psi))


########################################################
# Composition of operators
########################################################


@lru_cache(maxsize=int(1e6))
def com(
    A: Callable,
    B: Callable,
    psi: UncoupledBasisState,
    coefficients: HamiltonianConstants,
) -> UncoupledState:
    ABpsi = UncoupledState()
    # operate with A on all components in B|psi>
    for amp, cpt in B(psi, coefficients):
        ABpsi += amp * A(cpt, coefficients)
    return ABpsi



from __future__ import annotations

import copy
from collections.abc import Iterable
from typing import Any, List, Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt

from .constants import TlFNuclearSpins
from .find_states import QuantumSelector, get_unique_basisstates_from_basisstates
from .states import Basis, CoupledBasisState, ElectronicState, UncoupledBasisState
from .utils import parity_X

__all__ = [
    "generate_uncoupled_states_ground",
    "generate_uncoupled_states_excited",
    "generate_coupled_states_ground",
    "generate_coupled_states_excited",
    "generate_coupled_states_X",
    "generate_coupled_states_B",
]


def generate_uncoupled_states_ground(
    Js: Union[List[int], npt.NDArray[np.int_]],
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
) -> npt.NDArray[Any]:
    I_Tl = nuclear_spins.I_Tl
    I_F = nuclear_spins.I_F
    # convert J to int(J); np.int with (-1)**J throws an exception for negative J
    QN = np.array(
        [
            UncoupledBasisState(
                int(J),
                mJ,
                I_Tl,
                m1,
                I_F,
                m2,
                Omega=0,
                P=parity_X(J),
                electronic_state=ElectronicState.X,
                basis=Basis.Uncoupled,
            )
            for J in Js
            for mJ in range(-J, J + 1)
            for m1 in np.arange(-I_Tl, I_Tl + 1)
            for m2 in np.arange(-I_F, I_F + 1)
        ]
    )
    return QN


def generate_uncoupled_states_excited(
    Js: Union[List[int], npt.NDArray[np.int_]],
    Ωs: List[int] = [-1, 1],
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
) -> npt.NDArray[Any]:
    I_Tl = nuclear_spins.I_Tl
    I_F = nuclear_spins.I_F
    QN = np.array(
        [
            UncoupledBasisState(
                J,
                mJ,
                I_Tl,
                m1,
                I_F,
                m2,
                Omega=Ω,
                electronic_state=ElectronicState.B,
                basis=Basis.Uncoupled,
            )
            for Ω in Ωs
            for J in Js
            for mJ in range(-J, J + 1)
            for m1 in np.arange(-I_Tl, I_Tl + 1)
            for m2 in np.arange(-I_F, I_F + 1)
        ]
    )
    return QN


def generate_coupled_states_ground(
    Js: Union[List[int], npt.NDArray[np.int_]],
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
) -> npt.NDArray[Any]:
    I_Tl = nuclear_spins.I_Tl
    I_F = nuclear_spins.I_F
    QN = np.array(
        [
            CoupledBasisState(
                F,
                mF,
                F1,
                J,
                I_F,
                I_Tl,
                electronic_state=ElectronicState.X,
                P=parity_X(J),
                Omega=0,
                basis=Basis.Coupled,
            )
            for J in Js
            for F1 in np.arange(np.abs(J - I_F), J + I_F + 1)
            for F in np.arange(np.abs(F1 - I_Tl), F1 + I_Tl + 1)
            for mF in np.arange(-F, F + 1)
        ]
    )
    return QN


def generate_coupled_states_excited(
    Js: Union[List[int], npt.NDArray[np.int_]],
    Ps: Union[int, List[int], Tuple[int]] = 1,
    Omegas: Union[int, List[int], Tuple[int]] = 1,
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    basis: Optional[Basis] = None,
) -> npt.NDArray[Any]:
    I_Tl = nuclear_spins.I_Tl
    I_F = nuclear_spins.I_F

    if not isinstance(Ps, (list, tuple)):
        _Ps = [Ps]
    else:
        _Ps = list(Ps)
    if not isinstance(Omegas, (list, tuple)):
        _Omegas = [Omegas]
    else:
        _Omegas = list(Omegas)

    if len(_Omegas) > 1 and Ps is not None:
        raise ValueError("Cannot supply both Ω and P, need to pick a basis")
    elif Ps is not None and len(_Omegas) > 1:
        raise ValueError("Cannot supply both Ω and P, need to pick a basis")

    if basis == Basis.CoupledΩ:
        QN = np.array(
            [
                CoupledBasisState(
                    F,
                    mF,
                    F1,
                    J,
                    I_F,
                    I_Tl,
                    electronic_state=ElectronicState.B,
                    P=P,
                    Omega=Omega,
                    basis=basis,
                )
                for J in Js
                for F1 in np.arange(np.abs(J - I_F), J + I_F + 1)
                for F in np.arange(np.abs(F1 - I_Tl), F1 + I_Tl + 1)
                for mF in np.arange(-F, F + 1)
                for P in _Ps
                for Omega in _Omegas
            ]
        )
    else:
        QN = np.array(
            [
                CoupledBasisState(
                    F,
                    mF,
                    F1,
                    J,
                    I_F,
                    I_Tl,
                    electronic_state=ElectronicState.B,
                    P=P,
                    Omega=Omega,
                    basis=basis,
                )
                for J in Js
                for F1 in np.arange(np.abs(J - I_F), J + I_F + 1)
                for F in np.arange(np.abs(F1 - I_Tl), F1 + I_Tl + 1)
                for mF in np.arange(-F, F + 1)
                for P in _Ps
                for Omega in _Omegas
            ]
        )
    return QN


def generate_coupled_states_base(
    qn_selector: QuantumSelector,
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    basis: Optional[Basis] = None,
) -> List[CoupledBasisState]:
    """generate CoupledBasisStates for the quantum numbers given by qn_selector

    Args:
        qn_selector (QuantumSelector): quantum numbers to use to generate the
                                        CoupledBasisStates

    Returns:
        np.ndarray: array of CoupledBasisStates for the excited state
    """
    if (basis is not None) and (basis is not basis.CoupledΩ):
        assert qn_selector.P is not None, "function requires a parity to be set"
    assert (
        qn_selector.J is not None
    ), "function requires a rotational quantum number to be set"
    assert (
        qn_selector.electronic is not None
    ), "function requires electronic state to be set"
    assert qn_selector.Ω is not None, "function requires Ω to be set"

    # generate all combinations
    quantum_numbers = []
    for par in ["J", "F1", "F", "mF", "electronic", "P", "Ω"]:
        par_val = getattr(qn_selector, par)
        quantum_numbers.append(
            [par_val] if not isinstance(par_val, (list, tuple, np.ndarray)) else par_val
        )

    I_Tl = nuclear_spins.I_Tl
    I_F = nuclear_spins.I_F

    QN: List[CoupledBasisState] = []
    # the worst nested loops I've ever created
    Js, F1s, Fs, mFs, estates, Ps, Ωs = quantum_numbers
    for estate in estates:
        for J in Js:
            F1_allowed = np.arange(np.abs(J - I_F), J + I_F + 1)
            F1sl = F1s if F1s[0] is not None else F1_allowed
            for F1 in F1sl:
                if F1 not in F1_allowed:
                    continue
                Fs_allowed = np.arange(np.abs(F1 - I_Tl), F1 + I_Tl + 1)
                Fsl = Fs if Fs[0] is not None else Fs_allowed
                for F in Fsl:
                    if F not in Fs_allowed:
                        continue
                    mF_allowed = np.arange(-F, F + 1)
                    mFsl = mFs if mFs[0] is not None else mF_allowed
                    for mF in mFsl:
                        if mF not in mF_allowed:
                            continue
                        for P in Ps:
                            P = P if not callable(P) else P(J)
                            for Ω in Ωs:
                                QN.append(
                                    CoupledBasisState(
                                        F,
                                        mF,
                                        F1,
                                        J,
                                        I_F,
                                        I_Tl,
                                        electronic_state=estate,
                                        P=P,
                                        Ω=Ω,
                                        basis=basis,
                                    )
                                )
    return QN


def generate_coupled_states_X(
    qn_selector: Union[QuantumSelector, Sequence[QuantumSelector], npt.NDArray[Any]],
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    basis: Basis = Basis.Coupled,
) -> List[CoupledBasisState]:
    """generate ground X state CoupledBasisStates for the quantum numbers given
    by qn_selector

    Args:
        qn_selector (Union[QuantumSelector, list, np.ndarray]): QuantumSelector
            or list/array of QuantumSelectors to use for generating the
            CoupledBasisStates

    Returns:
        np.ndarray: array of CoupledBasisStates for the excited state
    """
    if isinstance(qn_selector, QuantumSelector):
        qns = copy.copy(qn_selector)
        qns.Ω = 0
        qns.P = parity_X
        qns.electronic = ElectronicState.X
        return generate_coupled_states_base(qns, nuclear_spins=nuclear_spins)
    elif isinstance(qn_selector, (list, np.ndarray)):
        coupled_states = []
        for qns in qn_selector:
            qns = copy.copy(qns)
            qns.Ω = 0
            qns.P = parity_X
            qns.electronic = ElectronicState.X
            coupled_states.append(
                generate_coupled_states_base(
                    qns, nuclear_spins=nuclear_spins, basis=basis
                )
            )

        return get_unique_basisstates_from_basisstates(
            [item for sublist in coupled_states for item in sublist]
        )
    else:
        raise AssertionError(
            "qn_selector required to be of type QuantumSelector, list or np.ndarray"
        )


def check_B_basis(P, Ω):
    if P is None and Ω is None:
        raise ValueError("Need to supply P and Ω to determine the basis")
    elif isinstance(Ω, Iterable):
        if isinstance(P, Iterable) and len(P) > 1:
            raise ValueError("Cannot supply both Ω and P, need to pick a basis")
        elif P is not None:
            raise ValueError("Cannot supply both Ω and P, need to pick a basis")
    else:
        return


def generate_coupled_states_B(
    qn_selector: Union[QuantumSelector, Sequence[QuantumSelector], npt.NDArray[Any]],
    nuclear_spins: TlFNuclearSpins = TlFNuclearSpins(),
    basis: Optional[Basis] = None,
) -> List[CoupledBasisState]:
    """generate excited B state CoupledBasisStates for the quantum numbers given
    by qn_selector

    Args:
        qn_selector (Union[QuantumSelector, list, np.ndarray]): QuantumSelector
            or list/array of QuantumSelectors to use for generating the
            CoupledBasisStates

    Returns:
        np.ndarray: array of CoupledBasisStates for the excited state
    """
    if isinstance(qn_selector, QuantumSelector):
        qns = copy.copy(qn_selector)
        qns.Ω = 1 if qns.Ω is None else qns.Ω
        qns.electronic = ElectronicState.B
        check_B_basis(qns.P, qns.Ω)
        return generate_coupled_states_base(qns, nuclear_spins=nuclear_spins)
    elif isinstance(qn_selector, (list, np.ndarray)):
        coupled_states = []
        for qns in qn_selector:
            qns = copy.copy(qns)
            qns.Ω = 1 if qns.Ω is None else qns.Ω
            qns.electronic = ElectronicState.B
            check_B_basis(qns.P, qns.Ω)
            coupled_states.append(
                generate_coupled_states_base(
                    qns, nuclear_spins=nuclear_spins, basis=basis
                )
            )
        return get_unique_basisstates_from_basisstates(
            [item for sublist in coupled_states for item in sublist]
        )
    else:
        raise AssertionError(
            "qn_selector required to be of type QuantumSelector, list or np.ndarray"
        )



import sympy as smp

__all__ = ["generate_density_matrix_symbolic"]


def recursive_subscript(i: int) -> str:
    # chr(0x2080+i) is unicode for
    # subscript num(i), resulting in x₀₀ for example
    if i < 10:
        return chr(0x2080 + i)
    else:
        return recursive_subscript(i // 10) + chr(0x2080 + i % 10)


def generate_density_matrix_symbolic(
    levels: int,
) -> smp.matrices.dense.MutableDenseMatrix:
    ρ = smp.zeros(levels, levels)
    levels = levels
    for i in range(levels):
        for j in range(i, levels):
            # \u03C1 is unicode for ρ,
            if i == j:
                ρ[i, j] = smp.Symbol(
                    "\u03C1{0},{1}".format(
                        recursive_subscript(i), recursive_subscript(j)
                    )
                )
            else:
                ρ[i, j] = smp.Symbol(
                    "\u03C1{0},{1}".format(
                        recursive_subscript(i), recursive_subscript(j)
                    )
                )
                ρ[j, i] = smp.Symbol(
                    "\u03C1{1},{0}".format(
                        recursive_subscript(i), recursive_subscript(j)
                    )
                )
    return ρ



from functools import lru_cache
from typing import Union

from sympy.physics.wigner import wigner_3j, wigner_6j

__all__ = ["threej_f", "sixj_f"]


@lru_cache(maxsize=int(1e6))
def threej_f(
    j1: Union[float, int],
    j2: Union[float, int],
    j3: Union[float, int],
    m1: Union[float, int],
    m2: Union[float, int],
    m3: Union[float, int],
) -> complex:
    return complex(wigner_3j(j1, j2, j3, m1, m2, m3))


@lru_cache(maxsize=int(1e6))
def sixj_f(
    j1: Union[float, int],
    j2: Union[float, int],
    j3: Union[float, int],
    j4: Union[float, int],
    j5: Union[float, int],
    j6: Union[float, int],
) -> complex:
    return complex(wigner_6j(j1, j2, j3, j4, j5, j6))



from typing import Sequence, Union, overload

import numpy as np
import numpy.typing as npt
import scipy.constants as cst

from centrex_tlf import states

__all__ = [
    "J_levels",
    "thermal_population",
    "generate_uniform_population_state_indices",
    "generate_uniform_population_states",
]


def J_levels(J: int) -> int:
    """
    Number of hyperfine levels per J rotational level

    Args:
        J (int): rotational level

    Returns:
        int: number of hyperfine levels
    """
    return 4 * (2 * J + 1)


@overload
def thermal_population(
    J: npt.NDArray[np.int_], T: float, B: float = 6.66733e9, n: int = 100
) -> npt.NDArray[np.float_]: ...


@overload
def thermal_population(
    J: int, T: float, B: float = 6.66733e9, n: int = 100
) -> float: ...


def thermal_population(J, T, B=6.66733e9, n=100):
    """
    Thermal population of a given J sublevel

    Args:
        J (int): Rotational level
        T (float): Temperature [Kelvin]
        B (float, optional): Rotational constant. Defaults to 6.66733e9.
        n (int, optional): Number of rotational levels to normalize with. Defaults to
                            100.

    Returns:
        float: relative population in a rational level
    """
    c = 2 * np.pi * cst.hbar * B / (cst.k * T)

    def a(J):
        return -c * J * (J + 1)

    Z = np.sum([J_levels(i) * np.exp(a(i)) for i in range(n)])
    return J_levels(J) * np.exp(a(J)) / Z


def generate_uniform_population_state_indices(
    state_indices: Sequence[int], levels: int
) -> npt.NDArray[np.complex_]:
    """
    spread population uniformly over the given state indices

    Args:
        state_indices (Sequence[int]): indices to put population into
        levels (int): total states involved

    Returns:
        npt.NDArray[np.complex_]: density matrix
    """
    ρ = np.zeros([levels, levels], dtype=complex)
    for ids in state_indices:
        ρ[ids, ids] = 1
    return ρ / np.trace(ρ)


def generate_uniform_population_states(
    selected_states: Union[
        Sequence[states.QuantumSelector],
        states.QuantumSelector,
    ],
    QN: Sequence[states.CoupledState],
) -> npt.NDArray[np.complex_]:
    """
    spread population uniformly over the given states

    Args:
        selected_states
            (Union[ Sequence[states.QuantumSelector], states.QuantumSelector, ]):
                state selector(s)
        QN (Sequence[states.State]): all states involved

    Returns:
        npt.NDArray[np.complex_]: density matrix
    """
    levels = len(QN)
    ρ = np.zeros([levels, levels], dtype=complex)

    if isinstance(selected_states, states.QuantumSelector):
        indices = selected_states.get_indices(QN)
    elif isinstance(selected_states, (list, tuple, np.ndarray)):
        indices = np.array([], dtype=np.int_)
        for ss in selected_states:
            indices = np.append(indices, ss.get_indices)
        indices = np.unique(indices)

    for ids in indices:
        ρ[ids, ids] = 1

    return ρ / np.trace(ρ)


def generate_thermal_population_states(
    temperature: float,
    QN: Sequence[states.CoupledState],
) -> npt.NDArray[np.complex_]:
    levels = len(QN)
    ρ = np.zeros([levels, levels], dtype=complex)

    assert isinstance(QN[0], states.CoupledState), "no State objects supplies"

    j_levels = np.unique([qn.largest.J for qn in QN])

    # get the relative thermal population fractions of the ground state
    population = dict(
        [(j, p) for j, p in zip(j_levels, thermal_population(j_levels, temperature))]
    )

    # get quantum numbers of the ground state
    quantum_numbers = [
        (qn.largest.J, qn.largest.F1, qn.largest.F, qn.largest.mF)
        for qn in QN
        if qn.largest.electronic_state == states.ElectronicState.X
    ]

    assert len(np.unique(quantum_numbers, axis=0)) == len(
        quantum_numbers
    ), "duplicate quantum numbers"

    for idx, qn in enumerate(QN):
        if qn.largest.F is None:
            ρ[idx, idx] = population[qn.largest.J]
        else:
            ρ[idx, idx] = population[qn.largest.J] / J_levels(qn.largest.J)

    return ρ



from typing import List, Sequence, cast, overload

import numpy as np
import numpy.typing as npt

from .rabi import intensity_to_rabi

__all__ = [
    "multipas_prism_order",
    "gaussian",
    "gaussian_amp",
    "gaussian_2d",
    "gaussian_2d_amp",
    "generate_2d_multipass_gaussian_intensity",
    "generate_2d_multipass_gaussian_rabi",
]


def multipas_prism_order(passes: int) -> List[int]:
    """
    Order of the multipass generated by two prisms.

    E.g. for 5 passes returns:
    [1,4,3,2,5]

    Args:
        passes (int): number of total passes

    Returns:
        List[int]: List with the order of the passes.
    """
    if passes % 2 == 0:
        raise ValueError("number of passes required to be odd")

    npass = [1]
    for p in range(1, passes):
        if p % 2 == 0:
            npass.append(p + 1)
        else:
            npass.append(passes - p)
    return npass


@overload
def gaussian(
    x: npt.NDArray[np.float_], mean: float, sigma: float
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian(x: float, mean: float, sigma: float) -> float: ...


def gaussian(x, mean, sigma):
    """
    non-normalized gaussian function

    Args:
        x (float): x value to evaluate gaussian at
        mean (float): mean of gaussian
        sigma (float): standard deviation of gaussian

    Returns:
        float: gaussian evaluated at x
    """
    return np.exp(-((x - mean) ** 2) / (2 * sigma**2))


@overload
def gaussian_amp(
    x: npt.NDArray[np.float_], a: float, mean: float, sigma: float
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian_amp(x: float, a: float, mean: float, sigma: float) -> float: ...


def gaussian_amp(x, a, mean, sigma):
    """
    Non-normalized gaussian function with amplitude a
    Args:
        x (float): x values to evaluate gaussian
        a (float): amplitude
        mean (float): mean of gaussian
        sigma (float): standard deviation of gaussian
    Returns:
        (float): gaussian evaluated at x
    """
    return a * gaussian(x, mean, sigma)


@overload
def gaussian_2d(
    x: npt.NDArray[np.float_],
    y: npt.NDArray[np.float_],
    mean_x: float,
    mean_y: float,
    sigma_x: float,
    sigma_y: float,
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian_2d(
    x: npt.NDArray[np.float_],
    y: npt.NDArray[np.float_],
    mean_x: npt.NDArray[np.float_],
    mean_y: npt.NDArray[np.float_],
    sigma_x: float,
    sigma_y: float,
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian_2d(
    x: float,
    y: float,
    mean_x: float,
    mean_y: float,
    sigma_x: float,
    sigma_y: float,
) -> float: ...


def gaussian_2d(x, y, mean_x, mean_y, sigma_x, sigma_y):
    """
    Non-normalized 2D gaussian function
    Args:
        x (float): x value to evaluate gaussian at
        y (float): y value to evaluate gaussian at
        mean_x (float): mean of gaussian in x
        mean_y (float): mean of gaussian in y
        sigma_x (float): standard deviation of gaussian in x
        sigma_y (float): standard deviation of gaussian in y
    Returns:
        (float): gaussian evaluated at x and y
    """
    a = (x - mean_x) ** 2 / (2 * sigma_x**2)
    b = (y - mean_y) ** 2 / (2 * sigma_y**2)
    return np.exp(-(a + b))


@overload
def gaussian_2d_amp(
    x: npt.NDArray[np.float_],
    y: npt.NDArray[np.float_],
    a: float,
    mean_x: float,
    mean_y: float,
    sigma_x: float,
    sigma_y: float,
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian_2d_amp(
    x: npt.NDArray[np.float_],
    y: npt.NDArray[np.float_],
    a: npt.NDArray[np.float_],
    mean_x: npt.NDArray[np.float_],
    mean_y: npt.NDArray[np.float_],
    sigma_x: float,
    sigma_y: float,
) -> npt.NDArray[np.float_]: ...


@overload
def gaussian_2d_amp(
    x: float,
    y: float,
    a: float,
    mean_x: float,
    mean_y: float,
    sigma_x: float,
    sigma_y: float,
) -> float: ...


def gaussian_2d_amp(
    x,
    y,
    a,
    mean_x,
    mean_y,
    sigma_x,
    sigma_y,
):
    """
    Non-normalized 2D gaussian function
    Args:
        x (float): x value to evaluate gaussian at
        y (float): y value to evaluate gaussian at
        a (float): amplitude
        mean_x (float): mean of gaussian in x
        mean_y (float): mean of gaussian in y
        sigma_x (float): standard deviation of gaussian in x
        sigma_y (float): standard deviation of gaussian in y
    Returns:
        (float): gaussian evaluated at x and y
    """
    return a * gaussian_2d(x, y, mean_x, mean_y, sigma_x, sigma_y)


def generate_2d_multipass_gaussian_intensity(
    X: npt.NDArray[np.float_],
    Y: npt.NDArray[np.float_],
    locations_x: Sequence[float],
    locations_y: Sequence[float],
    intensities: Sequence[float],
    sigma_x: float,
    sigma_y: float,
) -> npt.NDArray[np.float_]:
    """
    generate the intensity in 2D of a multipass

    Args:
        X (npt.NDArray[np.float_]): X values to evaluate multipass at
        Y (npt.NDArray[np.float_]): Y values to evaluate multipass at
        locations_x (Sequence[float]): x locations of passes
        locations_y (Sequence[float]): y locations of passes
        intensities (Sequence[float]): intensities of each pass
        sigma_x (float): standard deviation in x
        sigma_y (float): standard deviation in y

    Returns:
        npt.NDArray[np.float_]: intensity at X, Y coordinates
    """
    assert (
        len(locations_x) == len(locations_y) == len(intensities)
    ), "length of locations_x, locations_y and intensities must be equal"
    _intensities = cast(npt.NDArray[np.float_], np.asarray(intensities))
    _locations_x = cast(npt.NDArray[np.float_], np.asarray(locations_x))
    _locations_y = cast(npt.NDArray[np.float_], np.asarray(locations_y))
    multipass = gaussian_2d_amp(
        x=X[:, :, np.newaxis],
        y=Y[:, :, np.newaxis],
        a=_intensities[np.newaxis, :],
        mean_x=_locations_x[np.newaxis, :],
        mean_y=_locations_y[np.newaxis, :],
        sigma_x=sigma_x,
        sigma_y=sigma_y,
    ).sum(axis=2)
    return multipass


def generate_2d_multipass_gaussian_rabi(
    X: npt.NDArray[np.float_],
    Y: npt.NDArray[np.float_],
    locations_x: Sequence[float],
    locations_y: Sequence[float],
    intensities: Sequence[float],
    sigma_x: float,
    sigma_y: float,
    coupling: float,
    D: float = 2.6675506e-30,
) -> npt.NDArray[np.float_]:
    """
    Generate the rabi rate in 2D of a multipass.

    Args:
        X (npt.NDArray[np.float_]): X values to evaluate multipass at
        Y (npt.NDArray[np.float_]): Y values to evaluate multipass at
        locations_x (Sequence[float]): x locations of passes
        locations_y (Sequence[float]): y locations of passes
        intensities (Sequence[float]): intensities of each pass
        sigma_x (float): standard deviation in x
        sigma_y (float): standard deviation in y
        coupling (float): coupling strength
        D (float, optional): Dipole moment. Defaults to 2.6675506e-30 for the X to B TlF
                            transition.

    Returns:
        npt.NDArray[np.float_]: rabi rate at X, Y coordinates
    """
    multipass = generate_2d_multipass_gaussian_intensity(
        X, Y, locations_x, locations_y, intensities, sigma_x, sigma_y
    )
    rabi = intensity_to_rabi(intensity=multipass, coupling=coupling, D=D)
    return rabi



from functools import lru_cache

from centrex_tlf.states import CoupledBasisState, CoupledState

from ..constants import BConstants
from ..quantum_operators import J2, J4, J6


@lru_cache(maxsize=int(1e6))
def Hrot(psi: CoupledBasisState, constants: BConstants) -> CoupledState:
    """
    Rotational Hamiltonian for the B-state.
    """
    return (
        constants.B_rot * J2(psi)
        + constants.D_rot * J4(psi)
        + constants.H_const * J6(psi)
    )



from typing import Tuple

import numpy as np
import numpy.typing as npt
from scipy import special


def sideband_spectrum(
    β: float, ω: float, kmax: int
) -> Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    """
    Generate the sideband spectrum of a phase modulation eom, generated with:
    J0(β) + ΣJₖ(β) + Σ(-1)ᵏJₖ(β)
    summing over k from 0 to kmax, where the first sum is for peaks at positive detuning
    and the second sum for peaks at negative detuning.

    Args:
        kmax (int): maximum number of sidebands to plot

    Returns:
        Tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]: ndarray with frequencies
        and ndarray with sideband amplitudes
    """
    ks = np.arange(-kmax, kmax + 1, 1)
    ωs = ks * ω

    m = ks < 0
    ks[m] *= -1

    sidebands = special.jv(ks, β)
    sidebands[m] = sidebands[m] * (-1) ** ks[m]

    return ωs, sidebands
